# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import base64
from collections.abc import MutableSet as MutableSetBase, Set as SetBase
import json
from logging import Logger
import os
from pathlib import Path
import sys
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Collection,
    Iterator,
    Mapping,
    MutableSet,
    Optional,
    Union,
)

from ...prim import Party
from .exc import ConfigError

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

__all__ = [
    "AccessConfig",
    "TokenBasedAccessConfig",
    "PropertyBasedAccessConfig",
    "PartyRights",
    "PartyRightsSet",
    "create_access",
]

if TYPE_CHECKING:
    # We refer to the Config class in a docstring and
    # without this import, Sphinx can't resolve the reference
    # noinspection PyUnresolvedReferences
    from . import Config


def parties_from_env(*env_vars: str) -> AbstractSet[Party]:
    """
    Read the set of parties
    """
    return {Party(p) for env_var in env_vars for p in os.getenv(env_var, "").split(",") if p}


# mypy note: typing.overload cannot properly express a more correct signature for this function,
#  which is that if oauth_token is supplied, then a TokenBasedAccessConfig class is returned and
#  otherwise, a PropertyBasedAccessConfig class is returned. Trying to type this properly with
#  overloads results in:
#     "Not all union combinations were tried because there are too many unions"
#  It's also worth nothing that our only usage of this function also supplies all parameters since
#  we're effectively just passing **kwargs around, so it's not clear a more accurate type helps much
#  anyway.
def create_access(
    *,
    read_as: Union[None, Party, Collection[Party]] = None,
    act_as: Union[None, Party, Collection[Party]] = None,
    admin: Optional[bool] = None,
    ledger_id: Optional[str] = None,
    application_name: Optional[str] = None,
    oauth_token: Optional[str] = None,
    oauth_token_file: Optional[str] = None,
    logger: Optional[Logger] = None,
) -> "AccessConfig":
    """
    Create an appropriate instance of :class:`AccessConfig`.

    See :meth:`Config.create` for a more detailed description of these parameters.
    """
    # admin = None is effectively the same as admin = False in this context
    is_property_based = read_as or act_as or admin or ledger_id or application_name
    if not is_property_based and not oauth_token and not oauth_token_file:
        # none of the access-related parameters were passed in, so try to read some from the
        # environment
        act_as = parties_from_env("DAML_LEDGER_ACT_AS", "DAML_LEDGER_PARTY")
        read_as = parties_from_env("DAML_LEDGER_READ_AS", "DABL_PUBLIC_PARTY")
        ledger_id = os.getenv("DAML_LEDGER_ID", "")
        application_name = os.getenv("DAML_LEDGER_APPLICATION_NAME")
        oauth_token = os.getenv("DAML_LEDGER_OAUTH_TOKEN")
        oauth_token_file = os.getenv("DAML_LEDGER_OAUTH_TOKEN_FILE")

    is_property_based = read_as or act_as or admin or ledger_id or application_name
    if not is_property_based and not oauth_token and not oauth_token_file:
        raise ConfigError("no oauth token access or read_as/act_as/admin was specified")

    # how do they configure thee? let me count the ways...
    if sum(map(int, (bool(is_property_based), bool(oauth_token), bool(oauth_token_file)))) > 1:
        raise ConfigError(
            "must specify ONE of read_as/act_as/admin, oauth_token, or oauth_token_file"
        )

    if oauth_token_file:
        return TokenFileBasedAccessConfig(oauth_token_file)
    elif oauth_token:
        return TokenBasedAccessConfig(oauth_token)
    else:
        return PropertyBasedAccessConfig(
            read_as=read_as,
            act_as=act_as,
            admin=admin,
            ledger_id=ledger_id,
            application_name=application_name,
        )


@runtime_checkable
class AccessConfig(Protocol):
    """
    Configuration parameters for providing access to a ledger.

    To create an instance of this protocol, call :func:`create_access` and provide *either*
    ``oauth_token`` *or* the other fields you wish to set (such as ``act_as``). You cannot specify
    both an access token and other fields.

    You may implement this protocol using your own custom type if you have very specialized access
    needs.
    """

    @property
    def read_as(self) -> AbstractSet[Party]:
        """
        The set of parties that can be used to read data from the ledger. This also includes the
        set of parties that can be used to write data to the ledger.

        :type: AbstractSet[Party]
        """
        raise NotImplementedError

    @property
    def read_only_as(self) -> AbstractSet[Party]:
        """
        The set of parties that have read-only access to the underlying ledger.

        :type: AbstractSet[Party]
        """
        raise NotImplementedError

    @property
    def act_as(self) -> AbstractSet[Party]:
        """
        The set of parties that can be used to write data to the ledger.

        :type: AbstractSet[Party]
        """
        raise NotImplementedError

    @property
    def admin(self) -> bool:
        """
        ``True`` if the token grants "admin" access.

        :type: bool
        """
        raise NotImplementedError

    @property
    def ledger_id(self) -> Optional[str]:
        """
        The ledger ID. For non-token based access methods, this can be queried from the ledger.

        :type: Optional[str]
        """
        raise NotImplementedError

    @property
    def application_name(self) -> str:
        """
        The application name.

        :type: str
        """
        raise NotImplementedError

    @property
    def token(self) -> str:
        """
        The bearer token that provides authorization and authentication to a ledger.

        :type: str
        """
        raise NotImplementedError


class TokenBasedAccessConfig(AccessConfig):
    """
    Access configuration that is inherently token-based. The token can be changed at any time, and
    party rights, the application name, and ledger ID are all derived off of the token.
    """

    def __init__(self, oauth_token: str):
        """
        Initialize a token-based access configuration.

        :param oauth_token: The initial value of the bearer token.
        """
        self.token = oauth_token

    @property
    def token(self) -> str:
        """
        The bearer token that provides authorization and authentication to a ledger. This value can
        be replaced on a live connection in order to support use cases such as token refreshing.
        """
        return self._token

    @token.setter
    def token(self, value: str) -> None:
        self._token = value
        claims = decode_token(self._token)

        read_as = frozenset(claims.get("readAs", ()))
        act_as = frozenset(claims.get("actAs", ()))

        self._act_as = act_as
        self._read_only_as = read_as - act_as
        self._read_as = read_as.union(act_as)
        self._admin = bool(claims.get("admin", False))
        self._ledger_id = claims.get("ledgerId", None)
        self._application_name = claims.get("applicationId", None)

    @property
    def read_as(self) -> AbstractSet[Party]:
        return self._read_as

    @property
    def read_only_as(self) -> AbstractSet[Party]:
        return self._read_only_as

    @property
    def act_as(self) -> AbstractSet[Party]:
        return self._act_as

    @property
    def admin(self) -> bool:
        return self._admin

    @property
    def ledger_id(self) -> Optional[str]:
        return self._ledger_id

    @property
    def application_name(self) -> str:
        return self._application_name


class TokenFileBasedAccessConfig(TokenBasedAccessConfig):
    def __init__(self, oauth_token_file: str):
        # TODO: Update the token when the contents of this file change.
        super().__init__(Path(oauth_token_file).read_text())


class PropertyBasedAccessConfig(AccessConfig):
    """
    Access configuration that is manually specified outside of an authentication/authorization
    framework. Suitable for local testing or when no auth server is available, and the Ledger API
    inherently trusts any caller to provide its own authentication and authorization.
    """

    def __init__(
        self,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
        admin: Optional[bool] = False,
        ledger_id: Optional[str] = None,
        application_name: Optional[str] = None,
    ):
        """
        Initialize a property-based access configuration.

        :param read_as:
            A party or set of parties on whose behalf (in addition to all parties listed in ``act_as``)
            contracts can be retrieved.
        :param act_as:
            A party or set of parties on whose behalf commands should be executed. Parties here are also
            implicitly granted ``read_as`` access as well.
        :param admin:
            HTTP JSON API only: allow admin endpoints to be used. This flag is ignored when connecting
            to gRPC Ledger API implementations.
        :param ledger_id:
            The ledger ID to connect to. For the HTTP JSON API, this value is required. For the gRPC
            Ledger API, if this value is _not_ supplied, its value will be retrieved from the server.
        :param application_name:
            A string that identifies this application. This is used for tracing purposes on the
            server-side.
        """
        self._parties = PartyRights()
        self._parties.maybe_add(read_as, False)
        self._parties.maybe_add(act_as, True)
        self._admin = bool(admin)
        self._ledger_id = ledger_id
        self._application_name = application_name or "dazl-client"

    @property
    def token(self):
        """
        Produces a token without signing, utilizing our parameters.
        """
        return encode_unsigned_token(
            self.read_as, self.act_as, self.ledger_id, self.application_name, self.admin
        )

    @property
    def ledger_id(self) -> "Optional[str]":
        """
        The ledger ID. When connecting to the gRPC Ledger API, this can be inferred and does not
        need to be supplied. When connecting to the HTTP JSON API, it must be supplied.

        :type: Optional[str]
        """
        return self._ledger_id

    @ledger_id.setter
    def ledger_id(self, value: "Optional[str]") -> None:
        self._ledger_id = value

    @property
    def application_name(self) -> str:
        return self._application_name

    @property
    def read_as(self) -> "AbstractSet[Party]":
        """
        The set of parties for which read rights are granted. This collection is read-only. If you
        want to add a party with read-only access, add it to :meth:`read_only_as`; if you want to
        add a party with act-as access as well as read-only access, add it to :meth:`act_as`.

        This set always includes the act_as parties. For the set of parties that can be read as
        but NOT acted as, use :meth:`read_only_as`.

        :type: AbstractSet[Party]
        """
        return self._parties

    @property
    def read_only_as(self) -> "MutableSet[Party]":
        """
        The set of parties for which read-as rights are granted, but act-as rights are NOT granted.
        This collection can be modified.

        :type: MutableSet[Party]
        """
        return self._parties.read_as

    @property
    def act_as(self) -> "MutableSet[Party]":
        """
        The set of parties for which act-as rights are granted. This collection can be modified.
        Adding a party to this set _removes_ it from :meth:`read_only_as`.

        :type: MutableSet[Party]
        """
        return self._parties.act_as

    @property
    def admin(self) -> bool:
        """
        Whether or not the token sent to HTTP JSON API contains the ``admin: true`` flag that
        signals a token bearer with admin access.
        """
        return self._admin

    @admin.setter
    def admin(self, value: bool) -> None:
        self._admin = value


def parties(p: Union[None, Party, Collection[Party]]) -> Collection[Party]:
    if p is None:
        return []
    elif isinstance(p, str):
        return [Party(p)]
    else:
        return p


DamlLedgerApiNamespace = "https://daml.com/ledger-api"


def decode_token(token: str) -> Mapping[str, Any]:
    components = token.split(".", 3)
    if len(components) != 3:
        raise ValueError("not a JWT")
    claim_str = base64.urlsafe_b64decode(components[1])
    claims = json.loads(claim_str)
    claims_dict = claims.get(DamlLedgerApiNamespace)
    if claims_dict is None:
        raise ValueError(f"JWT is missing claim namespace: {DamlLedgerApiNamespace!r}")
    return claims_dict


def encode_unsigned_token(
    read_as: Collection[Party],
    act_as: Collection[Party],
    ledger_id: str,
    application_id: str,
    admin: bool = True,
) -> bytes:
    header = {
        "alg": "none",
        "typ": "JWT",
    }
    payload = {
        DamlLedgerApiNamespace: {
            "ledgerId": ledger_id,
            "applicationId": application_id,
            "actAs": sorted(act_as),
            "readAs": sorted(read_as),
            "admin": admin,
        }
    }

    return (
        base64.urlsafe_b64encode(json.dumps(header).encode("utf-8"))
        + b"."
        + base64.urlsafe_b64encode(json.dumps(payload).encode("utf-8"))
        + b"."
    )


class PartyRights(SetBase):
    __slots__ = ("_rights", "read_as", "act_as")

    def __init__(self):
        self._rights = dict()
        self.read_as = PartyRightsSet(self, False)
        self.act_as = PartyRightsSet(self, True)

    def maybe_add(
        self, value: "Union[None, Party, Collection[Party]]", has_act_rights: bool
    ) -> None:
        if value is None:
            return

        # Party is a fake Python newtype, so isinstance checks don't work on it
        if isinstance(value, str):
            self.add(Party(value), has_act_rights)
        else:
            for party in value:
                self.add(party, has_act_rights)

    def add(self, value: "Party", has_act_rights: bool) -> None:
        """
        Add/replace a ``Party`` and its rights.
        """
        self._rights[value] = has_act_rights

    def discard(self, value: "Party") -> None:
        self._rights.pop(value)

    def get(self, value: "Party") -> "Optional[bool]":
        return self._rights.get(value)

    def count(self, act_as: bool) -> int:
        return sum(1 for p, a in self._rights.items() if act_as == a)

    def __contains__(self, party: object) -> bool:
        return party in self._rights

    def __len__(self) -> int:
        return len(self._rights)

    def __iter__(self) -> "Iterator[Party]":
        return iter(sorted(self._rights))

    def iter(self, act_as: bool) -> "Iterator[Party]":
        return iter(p for p, a in sorted(self._rights.items()) if a == act_as)


class PartyRightsSet(MutableSetBase):
    def __init__(self, rights: "PartyRights", act_as: bool):
        self._rights = rights
        self._act_as = act_as

    def add(self, value: "Party") -> None:
        self._rights.add(value, self._act_as)

    def discard(self, value: "Party") -> None:
        self._rights.discard(value)

    def __contains__(self, obj: "object") -> bool:
        return isinstance(obj, str) and (self._rights.get(Party(obj)) == self._act_as)

    def __len__(self) -> int:
        return self._rights.count(self._act_as)

    def __iter__(self) -> "Iterator[Party]":
        return self._rights.iter(self._act_as)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Dict as __Map,
    List as __List,
    Literal as __Literal,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    overload,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.message import Message as __Message
from google.protobuf.wrappers_pb2 import StringValue

from .......prim import Party
from .value_pb2 import Identifier, Record, Value

__all__ = [
    "Event",
    "CreatedEvent",
    "ArchivedEvent",
    "ExercisedEvent",
]

class Event(__Message):
    created: CreatedEvent
    archived: ArchivedEvent
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, *, created: CreatedEvent = ...): ...
    @overload
    def __init__(self, *, archived: ArchivedEvent = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Event) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["event", "created", "archived"]) -> bool: ...
    def ClearField(self, field_name: __Literal["event", "created", "archived"]) -> None: ...
    def WhichOneof(self, oneof_group: __Literal["event"]) -> __Literal[None, "created", "archived"]: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class CreatedEvent(__Message):
    event_id: str
    contract_id: str
    template_id: Identifier
    contract_key: Value
    create_arguments: Record
    witness_parties: __List[Party]
    signatories: __List[Party]
    observers: __List[Party]
    agreement_text: StringValue
    def __init__(self, *, event_id: __Optional[str] = ..., contract_id: __Optional[str] = ..., template_id: __Optional[Identifier] = ..., contract_key: __Optional[Value] = ..., create_arguments: __Optional[Record] = ..., witness_parties: __Optional[__Sequence[Party]] = ..., signatories: __Optional[__Sequence[Party]] = ..., observers: __Optional[__Sequence[Party]] = ..., agreement_text: StringValue = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: CreatedEvent) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["event_id", "contract_id", "template_id", "contract_key", "create_arguments", "witness_parties", "signatories", "observers", "agreement_text"]) -> bool: ...
    def ClearField(self, field_name: __Literal["event_id", "contract_id", "template_id", "contract_key", "create_arguments", "witness_parties", "signatories", "observers", "agreement_text"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class ArchivedEvent(__Message):
    event_id: str
    contract_id: str
    template_id: Identifier
    witness_parties: __List[Party]
    def __init__(self, *, event_id: __Optional[str] = ..., contract_id: __Optional[str] = ..., template_id: __Optional[Identifier] = ..., witness_parties: __Optional[__Sequence[Party]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ArchivedEvent) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["event_id", "contract_id", "template_id", "witness_parties"]) -> bool: ...
    def ClearField(self, field_name: __Literal["event_id", "contract_id", "template_id", "witness_parties"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class ExercisedEvent(__Message):
    event_id: str
    contract_id: str
    template_id: Identifier
    choice: str
    choice_argument: Value
    acting_parties: __List[Party]
    consuming: bool
    witness_parties: __List[Party]
    child_event_ids: __List[str]
    exercise_result: Value
    def __init__(self, *, event_id: __Optional[str] = ..., contract_id: __Optional[str] = ..., template_id: __Optional[Identifier] = ..., choice: __Optional[str] = ..., choice_argument: __Optional[Value] = ..., acting_parties: __Optional[__Sequence[Party]] = ..., consuming: __Optional[bool] = ..., witness_parties: __Optional[__Sequence[Party]] = ..., child_event_ids: __Optional[__Sequence[str]] = ..., exercise_result: __Optional[Value] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ExercisedEvent) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["event_id", "contract_id", "template_id", "choice", "choice_argument", "acting_parties", "consuming", "witness_parties", "child_event_ids", "exercise_result"]) -> bool: ...
    def ClearField(self, field_name: __Literal["event_id", "contract_id", "template_id", "choice", "choice_argument", "acting_parties", "consuming", "witness_parties", "child_event_ids", "exercise_result"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    AsyncIterable as __AsyncStream,
    Awaitable as __Awaitable,
    Iterable as __Stream,
    Union as __Union,
    overload,
)

from grpc import Channel as __Channel
from grpc.aio import Channel as __AsyncChannel

from .ledger_identity_service_pb2 import GetLedgerIdentityRequest, GetLedgerIdentityResponse

__all__ = [
    "LedgerIdentityServiceStub",
]

class LedgerIdentityServiceStub:
    @classmethod
    @overload
    def __new__(cls, channel: __Channel) -> _LedgerIdentityServiceStub: ...
    @classmethod
    @overload
    def __new__(cls, channel: __AsyncChannel) -> _LedgerIdentityServiceStub_Async: ...
    def GetLedgerIdentity(self, __1: GetLedgerIdentityRequest) -> __Union[GetLedgerIdentityResponse, __Awaitable[GetLedgerIdentityResponse]]: ...

class _LedgerIdentityServiceStub(LedgerIdentityServiceStub):
    def GetLedgerIdentity(self, __1: GetLedgerIdentityRequest) -> GetLedgerIdentityResponse: ...

class _LedgerIdentityServiceStub_Async(LedgerIdentityServiceStub):
    def GetLedgerIdentity(self, __1: GetLedgerIdentityRequest) -> __Awaitable[GetLedgerIdentityResponse]: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/rpc/status.proto

from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="google/rpc/status.proto",
    package="google.rpc",
    syntax="proto3",
    serialized_options=b"\n\016com.google.rpcB\013StatusProtoP\001Z7google.golang.org/genproto/googleapis/rpc/status;status\370\001\001\242\002\003RPC",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x17google/rpc/status.proto\x12\ngoogle.rpc\x1a\x19google/protobuf/any.proto"N\n\x06Status\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12%\n\x07\x64\x65tails\x18\x03 \x03(\x0b\x32\x14.google.protobuf.AnyBa\n\x0e\x63om.google.rpcB\x0bStatusProtoP\x01Z7google.golang.org/genproto/googleapis/rpc/status;status\xf8\x01\x01\xa2\x02\x03RPCb\x06proto3',
    dependencies=[
        google_dot_protobuf_dot_any__pb2.DESCRIPTOR,
    ],
)


_STATUS = _descriptor.Descriptor(
    name="Status",
    full_name="google.rpc.Status",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="code",
            full_name="google.rpc.Status.code",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="message",
            full_name="google.rpc.Status.message",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="details",
            full_name="google.rpc.Status.details",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=66,
    serialized_end=144,
)

_STATUS.fields_by_name["details"].message_type = google_dot_protobuf_dot_any__pb2._ANY
DESCRIPTOR.message_types_by_name["Status"] = _STATUS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Status = _reflection.GeneratedProtocolMessageType(
    "Status",
    (_message.Message,),
    {
        "DESCRIPTOR": _STATUS,
        "__module__": "google.rpc.status_pb2"
        # @@protoc_insertion_point(class_scope:google.rpc.Status)
    },
)
_sym_db.RegisterMessage(Status)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains miscellaneous utility methods that don't really fit anywhere else.
"""

from typing import (
    Callable,
    Collection,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
import warnings

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


__all__ = ["boundary_iter", "flatten", "as_list", "get_matches"]


def boundary_iter(obj: Iterable[T]) -> Generator[Tuple[bool, T], None, None]:
    """
    Iterates over an iterable, returning a boolean and the underlying values indicating
    whether the iterator has reached its last item.

    >>> list(boundary_iter(range(0, 3)))
    [(False, 0), (False, 1), (True, 2)]

    >>> list(boundary_iter('A'))
    [(True, 'A')]

    # empty example
    >>> list(boundary_iter([]))
    []

    # iterating over an infinite iterable
    >>> import itertools
    >>> list(itertools.islice(boundary_iter(iter(lambda:0,1)), 4))
    [(False, 0), (False, 0), (False, 0), (False, 0)]
    """
    gen = iter(obj)
    prev_entry = None
    while True:
        try:
            entry = next(gen)
            if prev_entry is not None:
                yield False, prev_entry
            prev_entry = entry
        except StopIteration:
            if prev_entry is not None:
                yield True, prev_entry
            break


def flatten(obj):
    ret = []
    for sublist in obj:
        ret.extend(sublist)
    return ret


def as_list(obj: "Union[None, T, Collection[Union[None, T]]]") -> "List[T]":
    """
    Convert an object that is either nothing, a single object, or a collection, to a list of type
    of that object.
    """
    from collections.abc import Iterable

    if obj is None:
        return []
    elif isinstance(obj, str):
        # strings are iterable, but it's almost never intended to be used in place of a list; if
        # we're given a string, then assume what is wanted is a single list containing the full
        # string instead of a list containing every character as an individual item
        return [obj]  # type: ignore
    elif isinstance(obj, Iterable):
        return [o for o in obj if o is not None]
    else:
        # assume we're dealing with a single object of the requested type
        return [obj]


def __key_error(key: str) -> Exception:
    return KeyError(key)


def get_matches(
    mapping: "Mapping[str, V]",
    key: "str",
    exc_class: "Optional[Callable[[str], Exception]]" = __key_error,
) -> "Collection[V]":
    """
    Return the value (as a singleton collection) associated with a key. If the key is equal to the
    special value ``"*"``, then all values are returned. If there is no mapping for the specified
    key and an exception class is provided, an exception is raised. Otherwise, an empty collection
    is returned.

    This function should NOT be used if ``"*"`` _can_ be a valid key; the behavior in this case is
    still to return all values, which could be confusing.

    :param mapping:
        A mapping to use to perform lookups.
    :param key:
        The key to lookup a value for, or the special value '*' to retrieve all values.
    :param exc_class:
        The exception to raise if this function would have returned an empty collection, or
        ``None`` to allow empty collections to be returned.
    :return:
        * An empty collection if the supplied key does not have a mapping _and_ ``exc_class is not None``;
        * A collection of one if a match is found for the supplied key;
        * All values of the mapping if the key is `""*""`.
    """
    warnings.warn("get_matches is deprecated; there is no replacement.", DeprecationWarning)
    if key == "*":
        return tuple(mapping.values())

    value = mapping.get(key)
    if value is None:
        if exc_class is not None:
            raise exc_class(key)
        else:
            return ()

    return (value,)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/event.proto

from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2

from . import value_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="com/daml/ledger/api/v1/event.proto",
    package="com.daml.ledger.api.v1",
    syntax="proto3",
    serialized_options=b"\n\026com.daml.ledger.api.v1B\017EventOuterClass\252\002\026Com.Daml.Ledger.Api.V1",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n"com/daml/ledger/api/v1/event.proto\x12\x16\x63om.daml.ledger.api.v1\x1a"com/daml/ledger/api/v1/value.proto\x1a\x1egoogle/protobuf/wrappers.proto"\x95\x01\n\x05\x45vent\x12\x37\n\x07\x63reated\x18\x01 \x01(\x0b\x32$.com.daml.ledger.api.v1.CreatedEventH\x00\x12\x39\n\x08\x61rchived\x18\x03 \x01(\x0b\x32%.com.daml.ledger.api.v1.ArchivedEventH\x00\x42\x07\n\x05\x65ventJ\x04\x08\x02\x10\x03R\texercised"\xd4\x02\n\x0c\x43reatedEvent\x12\x10\n\x08\x65vent_id\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontract_id\x18\x02 \x01(\t\x12\x37\n\x0btemplate_id\x18\x03 \x01(\x0b\x32".com.daml.ledger.api.v1.Identifier\x12\x33\n\x0c\x63ontract_key\x18\x07 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.Value\x12\x38\n\x10\x63reate_arguments\x18\x04 \x01(\x0b\x32\x1e.com.daml.ledger.api.v1.Record\x12\x17\n\x0fwitness_parties\x18\x05 \x03(\t\x12\x13\n\x0bsignatories\x18\x08 \x03(\t\x12\x11\n\tobservers\x18\t \x03(\t\x12\x34\n\x0e\x61greement_text\x18\x06 \x01(\x0b\x32\x1c.google.protobuf.StringValue"\x88\x01\n\rArchivedEvent\x12\x10\n\x08\x65vent_id\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontract_id\x18\x02 \x01(\t\x12\x37\n\x0btemplate_id\x18\x03 \x01(\x0b\x32".com.daml.ledger.api.v1.Identifier\x12\x17\n\x0fwitness_parties\x18\x04 \x03(\t"\xd9\x02\n\x0e\x45xercisedEvent\x12\x10\n\x08\x65vent_id\x18\x01 \x01(\t\x12\x13\n\x0b\x63ontract_id\x18\x02 \x01(\t\x12\x37\n\x0btemplate_id\x18\x03 \x01(\x0b\x32".com.daml.ledger.api.v1.Identifier\x12\x0e\n\x06\x63hoice\x18\x05 \x01(\t\x12\x36\n\x0f\x63hoice_argument\x18\x06 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.Value\x12\x16\n\x0e\x61\x63ting_parties\x18\x07 \x03(\t\x12\x11\n\tconsuming\x18\x08 \x01(\x08\x12\x17\n\x0fwitness_parties\x18\n \x03(\t\x12\x17\n\x0f\x63hild_event_ids\x18\x0b \x03(\t\x12\x36\n\x0f\x65xercise_result\x18\x0c \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueJ\x04\x08\x04\x10\x05J\x04\x08\t\x10\nBB\n\x16\x63om.daml.ledger.api.v1B\x0f\x45ventOuterClass\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3',
    dependencies=[
        com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2.DESCRIPTOR,
        google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,
    ],
)


_EVENT = _descriptor.Descriptor(
    name="Event",
    full_name="com.daml.ledger.api.v1.Event",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="created",
            full_name="com.daml.ledger.api.v1.Event.created",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="archived",
            full_name="com.daml.ledger.api.v1.Event.archived",
            index=1,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="event",
            full_name="com.daml.ledger.api.v1.Event.event",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=131,
    serialized_end=280,
)


_CREATEDEVENT = _descriptor.Descriptor(
    name="CreatedEvent",
    full_name="com.daml.ledger.api.v1.CreatedEvent",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="event_id",
            full_name="com.daml.ledger.api.v1.CreatedEvent.event_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="contract_id",
            full_name="com.daml.ledger.api.v1.CreatedEvent.contract_id",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="template_id",
            full_name="com.daml.ledger.api.v1.CreatedEvent.template_id",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="contract_key",
            full_name="com.daml.ledger.api.v1.CreatedEvent.contract_key",
            index=3,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="create_arguments",
            full_name="com.daml.ledger.api.v1.CreatedEvent.create_arguments",
            index=4,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="witness_parties",
            full_name="com.daml.ledger.api.v1.CreatedEvent.witness_parties",
            index=5,
            number=5,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="signatories",
            full_name="com.daml.ledger.api.v1.CreatedEvent.signatories",
            index=6,
            number=8,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="observers",
            full_name="com.daml.ledger.api.v1.CreatedEvent.observers",
            index=7,
            number=9,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="agreement_text",
            full_name="com.daml.ledger.api.v1.CreatedEvent.agreement_text",
            index=8,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=283,
    serialized_end=623,
)


_ARCHIVEDEVENT = _descriptor.Descriptor(
    name="ArchivedEvent",
    full_name="com.daml.ledger.api.v1.ArchivedEvent",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="event_id",
            full_name="com.daml.ledger.api.v1.ArchivedEvent.event_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="contract_id",
            full_name="com.daml.ledger.api.v1.ArchivedEvent.contract_id",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="template_id",
            full_name="com.daml.ledger.api.v1.ArchivedEvent.template_id",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="witness_parties",
            full_name="com.daml.ledger.api.v1.ArchivedEvent.witness_parties",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=626,
    serialized_end=762,
)


_EXERCISEDEVENT = _descriptor.Descriptor(
    name="ExercisedEvent",
    full_name="com.daml.ledger.api.v1.ExercisedEvent",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="event_id",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.event_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="contract_id",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.contract_id",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="template_id",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.template_id",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="choice",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.choice",
            index=3,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="choice_argument",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.choice_argument",
            index=4,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="acting_parties",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.acting_parties",
            index=5,
            number=7,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="consuming",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.consuming",
            index=6,
            number=8,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="witness_parties",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.witness_parties",
            index=7,
            number=10,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="child_event_ids",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.child_event_ids",
            index=8,
            number=11,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="exercise_result",
            full_name="com.daml.ledger.api.v1.ExercisedEvent.exercise_result",
            index=9,
            number=12,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=765,
    serialized_end=1110,
)

_EVENT.fields_by_name["created"].message_type = _CREATEDEVENT
_EVENT.fields_by_name["archived"].message_type = _ARCHIVEDEVENT
_EVENT.oneofs_by_name["event"].fields.append(_EVENT.fields_by_name["created"])
_EVENT.fields_by_name["created"].containing_oneof = _EVENT.oneofs_by_name["event"]
_EVENT.oneofs_by_name["event"].fields.append(_EVENT.fields_by_name["archived"])
_EVENT.fields_by_name["archived"].containing_oneof = _EVENT.oneofs_by_name["event"]
_CREATEDEVENT.fields_by_name[
    "template_id"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._IDENTIFIER
_CREATEDEVENT.fields_by_name[
    "contract_key"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._VALUE
_CREATEDEVENT.fields_by_name[
    "create_arguments"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._RECORD
_CREATEDEVENT.fields_by_name[
    "agreement_text"
].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_ARCHIVEDEVENT.fields_by_name[
    "template_id"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._IDENTIFIER
_EXERCISEDEVENT.fields_by_name[
    "template_id"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._IDENTIFIER
_EXERCISEDEVENT.fields_by_name[
    "choice_argument"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._VALUE
_EXERCISEDEVENT.fields_by_name[
    "exercise_result"
].message_type = com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2._VALUE
DESCRIPTOR.message_types_by_name["Event"] = _EVENT
DESCRIPTOR.message_types_by_name["CreatedEvent"] = _CREATEDEVENT
DESCRIPTOR.message_types_by_name["ArchivedEvent"] = _ARCHIVEDEVENT
DESCRIPTOR.message_types_by_name["ExercisedEvent"] = _EXERCISEDEVENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Event = _reflection.GeneratedProtocolMessageType(
    "Event",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENT,
        "__module__": "com.daml.ledger.api.v1.event_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Event)
    },
)
_sym_db.RegisterMessage(Event)

CreatedEvent = _reflection.GeneratedProtocolMessageType(
    "CreatedEvent",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEDEVENT,
        "__module__": "com.daml.ledger.api.v1.event_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CreatedEvent)
    },
)
_sym_db.RegisterMessage(CreatedEvent)

ArchivedEvent = _reflection.GeneratedProtocolMessageType(
    "ArchivedEvent",
    (_message.Message,),
    {
        "DESCRIPTOR": _ARCHIVEDEVENT,
        "__module__": "com.daml.ledger.api.v1.event_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ArchivedEvent)
    },
)
_sym_db.RegisterMessage(ArchivedEvent)

ExercisedEvent = _reflection.GeneratedProtocolMessageType(
    "ExercisedEvent",
    (_message.Message,),
    {
        "DESCRIPTOR": _EXERCISEDEVENT,
        "__module__": "com.daml.ledger.api.v1.event_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExercisedEvent)
    },
)
_sym_db.RegisterMessage(ExercisedEvent)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    ClassVar as __ClassVar,
    Dict as __Map,
    List as __List,
    Literal as __Literal,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    overload,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.internal.enum_type_wrapper import EnumTypeWrapper as __EnumTypeWrapper
from google.protobuf.message import Message as __Message

__all__ = [
    "LedgerOffset",
]

class LedgerOffset(__Message):
    absolute: str
    boundary: __Literal[0, 1]
    def __init__(self, *, absolute: __Optional[str] = ..., boundary: __Optional[__Literal[0, 1]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: LedgerOffset) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["value", "absolute", "boundary"]) -> bool: ...
    def ClearField(self, field_name: __Literal["value", "absolute", "boundary"]) -> None: ...
    def WhichOneof(self, oneof_group: __Literal["value"]) -> __Literal[None, "absolute", "boundary"]: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

    class LedgerBoundary(__EnumTypeWrapper):
        @overload
        def Name(self, number: __Literal[0]) -> __Literal["LEDGER_BEGIN"]: ...
        @overload
        def Name(self, number: __Literal[1]) -> __Literal["LEDGER_END"]: ...
        @overload
        def Value(self, name: __Literal["LEDGER_BEGIN"]) -> __Literal[0]: ...
        @overload
        def Value(self, name: __Literal["LEDGER_END"]) -> __Literal[1]: ...

    LEDGER_BEGIN: __ClassVar[__Literal[0]] = ...
    LEDGER_END: __ClassVar[__Literal[1]] = ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import warnings

from ..ledger.pkgloader_aio_compat import PackageLoader, SyncPackageService

warnings.warn(
    "dazl.client.pkg_loader is deprecated; use dazl.protocols.pkgloader_aio instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = ["SyncPackageService", "PackageLoader"]



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Collection, Union

from ..prim import DazlError, DazlWarning, Party

__all__ = ["ConfigurationError", "DazlPartyMissingError", "UnknownTemplateWarning"]


class ConfigurationError(DazlError):
    """
    Raised when a configuration error prevents a client from being started.

    .. attribute:: ConfigurationError.reasons

        A collection of reasons for a failure.
    """

    reasons: Collection[str]

    def __init__(self, reasons: Union[None, str, Collection[str]]):
        if reasons is None:
            self.reasons = []
        elif isinstance(reasons, str):
            self.reasons = [reasons]
        else:
            self.reasons = reasons


class DazlPartyMissingError(DazlError):
    """
    Error raised when a party or some information about a party is requested, and that party is not
    found.
    """

    def __init__(self, party: "Party"):
        super().__init__(f"party {party!r} does not have a defined client")
        self.party = party


class UnknownTemplateWarning(DazlWarning):
    """
    Raised when trying to do something with a template name that is unknown.
    """



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import (
    participant_pruning_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_participant__pruning__service__pb2,
)


class ParticipantPruningServiceStub(object):
    """Status: experimental interface, will change before it is deemed production
    ready

    Prunes/truncates the "oldest" transactions from the participant (the participant Ledger Api Server plus any other
    participant-local state) by removing a portion of the ledger in such a way that the set of future, allowed
    commands are not affected.

    This enables:
    1. keeping the "inactive" portion of the ledger to a manageable size and
    2. removing inactive state to honor the right to be forgotten.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Prune = channel.unary_unary(
            "/com.daml.ledger.api.v1.admin.ParticipantPruningService/Prune",
            request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_participant__pruning__service__pb2.PruneRequest.SerializeToString,
            response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_participant__pruning__service__pb2.PruneResponse.FromString,
        )


class ParticipantPruningServiceServicer(object):
    """Status: experimental interface, will change before it is deemed production
    ready

    Prunes/truncates the "oldest" transactions from the participant (the participant Ledger Api Server plus any other
    participant-local state) by removing a portion of the ledger in such a way that the set of future, allowed
    commands are not affected.

    This enables:
    1. keeping the "inactive" portion of the ledger to a manageable size and
    2. removing inactive state to honor the right to be forgotten.
    """

    def Prune(self, request, context):
        """Prune the ledger specifying the offset before and at which ledger transactions should be removed. Only returns when
        the potentially long-running prune request ends successfully or with one of the following errors:
        - ``INVALID_ARGUMENT``: if the payload, particularly the offset is malformed or missing
        - ``UNIMPLEMENTED``: if the participant is based on a ledger that has not implemented pruning
        - ``INTERNAL``: if the participant has encountered a failure and has potentially applied pruning partially. Such cases
        warrant verifying the participant health before retrying the prune with the same (or a larger, valid) offset.
        Successful retries after such errors ensure that different components reach a consistent pruning state.

        Other GRPC errors can be returned depending on the type of condition preventing a prune:
        - ``OUT_OF_RANGE``: if the participant is not yet able to prune at the specified offset, but without user intervention
        the offset will eventually be usable for pruning.
        - ``FAILED_PRECONDITION`` if some sort of user intervention is required before pruning can proceed at the specified
        offset.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_ParticipantPruningServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Prune": grpc.unary_unary_rpc_method_handler(
            servicer.Prune,
            request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_participant__pruning__service__pb2.PruneRequest.FromString,
            response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_participant__pruning__service__pb2.PruneResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "com.daml.ledger.api.v1.admin.ParticipantPruningService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class ParticipantPruningService(object):
    """Status: experimental interface, will change before it is deemed production
    ready

    Prunes/truncates the "oldest" transactions from the participant (the participant Ledger Api Server plus any other
    participant-local state) by removing a portion of the ledger in such a way that the set of future, allowed
    commands are not affected.

    This enables:
    1. keeping the "inactive" portion of the ledger to a manageable size and
    2. removing inactive state to honor the right to be forgotten.
    """

    @staticmethod
    def Prune(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/com.daml.ledger.api.v1.admin.ParticipantPruningService/Prune",
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_participant__pruning__service__pb2.PruneRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_admin_dot_participant__pruning__service__pb2.PruneResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module has been relocated to ``dazl.client.events``, though if possible you should move to
``dazl.protocol.events``.
"""

from ..client._reader_sync import max_offset
from ..client.events import EventKey, create_dispatch
from ..protocols.events import (
    ActiveContractSetEvent,
    BaseEvent,
    BaseTransactionEvent,
    ContractArchiveEvent,
    ContractCreateEvent,
    ContractEvent,
    ContractExercisedEvent,
    ContractFilter,
    InitEvent,
    OffsetEvent,
    PackagesAddedEvent,
    ReadyEvent,
    TransactionEndEvent,
    TransactionFilter,
    TransactionStartEvent,
)

__all__ = [
    "ActiveContractSetEvent",
    "BaseEvent",
    "BaseTransactionEvent",
    "ContractArchiveEvent",
    "ContractCreateEvent",
    "ContractEvent",
    "ContractExercisedEvent",
    "ContractFilter",
    "EventKey",
    "InitEvent",
    "OffsetEvent",
    "PackagesAddedEvent",
    "ReadyEvent",
    "TransactionEndEvent",
    "TransactionFilter",
    "TransactionStartEvent",
    "create_dispatch",
    "max_offset",
]



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Utilities for dealing with the file system.
"""

from io import BufferedIOBase
from pathlib import Path
import socket
from typing import BinaryIO, Optional, Union, overload


@overload
def get_bytes(_: None) -> None:
    ...


@overload
def get_bytes(_: "Union[bytes, str, Path, BinaryIO]") -> bytes:
    ...


def get_bytes(src):
    """
    Read the contents of a file as ``bytes``. If ``src`` is ``None``, then ``None`` is returned
    instead.
    """
    if src is None:
        return None
    elif isinstance(src, str):
        with open(src, "rb") as f:
            return f.read()
    elif isinstance(src, bytes):
        return src
    elif isinstance(src, Path):
        return src.read_bytes()
    elif isinstance(src, BufferedIOBase):
        return src.read(-1)
    else:
        raise ValueError("src must be a string path, a Path, or a BytesIO")


def read_file_bytes(file_path: Optional[str]) -> Optional[bytes]:
    if file_path is not None:
        with open(file_path, "rb") as f:
            return f.read()
    else:
        return None


def find_free_port() -> int:
    sock = socket.socket()
    try:
        sock.bind(("", 0))
        return sock.getsockname()[1]
    finally:
        sock.close()


def is_port_alive(port: int) -> bool:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        result = sock.connect_ex(("127.0.0.1", port))
        return result == 0
    finally:
        sock.close()


def find_nearest_ancestor(file_name: str, relative_to: Union[str, Path]) -> "Optional[Path]":
    """
    Return a path to a file that has the specified file name, or ``None`` if no ancestor could be
    found.

    :param file_name: The name of the file to search for.
    :param relative_to: The directory at which to begin the search.
    :return: A :class:`Path` that matches the specified file name.
    """
    if relative_to is None:
        raise ValueError("A relative_to path is required")

    relative_to = Path(relative_to)
    if relative_to.is_file():
        relative_to = relative_to.parent

    while True:
        candidate = relative_to / file_name
        if candidate.exists():
            return candidate
        else:
            new_relative_to = relative_to.parent
            if relative_to == new_relative_to:
                # We can't walk up the tree any more, so we couldn't find the file we were looking for
                return None
            else:
                relative_to = new_relative_to



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
:mod:`dazl.ledger.config` — connection configuration
====================================================

This module contains configuration objects for a :class:`Connection`.

Normally you don't need to construct these objects directly; instead, simply call
:meth:`dazl.connect`, which contains the same options exposed by the :meth:`Config.create` function,
which in turn includes the options exposed by the subobjects of :class:`Config`.

Generally you should not modify a :class:`Config` object (or subobjects) that have already been
passed to a :class:`Connection`. There are some exceptions to this rule, though; for example, when
using tokens, you can simply assign ``Config.access.token`` to a new value.

Configuration options are broken up into three subobjects:

* :ref:`Access configuration <access-configuration>`: settings commonly found in Daml JWT tokens
  (party settings, ledger ID, application name)
* :ref:`SSL/TLS configuration <ssl-tls-configuration>`: settings for configuring TLS connections
* :ref:`URL Configuration <url-configuration>`: settings that determine the location of gRPC Ledger
  API or HTTP JSON API implementation

.. autoclass:: Config
    :members:

.. _access-configuration:

Access configuration
--------------------

The :class:`AccessConfig` protocol specifies how ``dazl`` identifies itself to a ledger. There are
two built-in mechanisms for this: **property-based access**, which is traditionally used with
ledgers that do NOT require authorization/authentication (typical in a local development scenario,
for example coding against a local sandbox), and **token-based access**, which is required for
ledgers that DO require authorization/authentication (typical in a production scenario and/or hosted
ledgers).

In **property-based access** (:class:`PropertyBasedAccessConfig`), the behavior differs depending on
whether you are connecting over the gRPC Ledger API or the HTTP JSON API:

* For the *gRPC Ledger API*, ``read_as`` and ``act_as`` are used as-is. ``ledger_id`` is
  defaulted to the value requested from the ledger (but only if not initially specified).
  The ``admin`` property is ignored and unused.
* For the *HTTP JSON API*, ``read_as``, ``act_as``, ``admin``, ``ledger_id``, and
  ``application_name`` are all used to generate an unsigned JWT locally. ``ledger_id`` MUST be
  supplied.

In **token-based access** (:class:`TokenBasedAccessConfig`), the value of the token completely
determines the parties that can be used, the ledger ID to connect to, and the name of the
application. ``AccessConfig.token`` can be overwritten at any time, and that value will be used for
all subsequent calls to the ledger. If your ledger requires authorization/authentication using
tokens, you _must_ use token-based access.

Although ``dazl`` does not currently refresh tokens automatically, you can update the token yourself
at any time:

.. code-block:: python

    async def main():
        async with dazl.connect(token=MY_INITIAL_TOKEN) as conn:
            task1 = asyncio.create_task(do_ledger_stuff(conn))
            task2 = asyncio.create_task(refresh_token(conn))

            await task1
            task2.cancel()

    async def do_ledger_stuff(conn):
        # use the connection normally to make ledger calls
        ...

    async def refresh(conn):
        while True:
            # sleep for one hour
            await asyncio.sleep(3600)
            conn.config.access.token = NEW_REFRESHED_TOKEN

Deeper support for token refreshing may be added in a future release.

.. autoclass:: AccessConfig()
    :members:

.. autofunction:: create_access

.. autoclass:: PropertyBasedAccessConfig()
    :members:

.. autoclass:: TokenBasedAccessConfig()
    :members:

.. _ssl-tls-configuration:

SSL/TLS configuration
---------------------

.. autoclass:: SSLConfig
    :members:

.. _url-configuration:

URL configuration
-----------------

The :class:`URLConfig` protocol specifies the ledger implementation that ``dazl`` connects to
(either gRPC Ledger API or HTTP JSON API).

.. autofunction:: create_url

.. autoclass:: URLConfig()
    :members:

"""

import itertools
import logging
from logging import Logger
from os import PathLike
from typing import Collection, Optional, Union

from ...prim import Party, TimeDeltaLike
from .access import AccessConfig, PropertyBasedAccessConfig, TokenBasedAccessConfig, create_access
from .argv import configure_parser
from .ssl import SSLConfig
from .url import URLConfig, create_url

__all__ = [
    "Config",
    "AccessConfig",
    "SSLConfig",
    "URLConfig",
    "TokenBasedAccessConfig",
    "PropertyBasedAccessConfig",
    "create_access",
    "create_url",
    "configure_parser",
]

# an incrementing counter that helps keep loggers for individual config connections unique
# (see https://mail.python.org/pipermail//python-ideas/2016-August/041871.html); this is safe to
# do even in multithreaded environments because ultimately we're reusing the GIL as our lock
id_generator = itertools.count()


# PyCharm thinks ``from .url import ...`` clashes with variables named ``url``
# (same with ``ssl`` and ``access``).
# noinspection PyShadowingNames
class Config:
    """
    Stores configuration for a :class:`Connection`.
    """

    @classmethod
    def create(
        cls,
        url: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        scheme: Optional[str] = None,
        read_as: Union[None, Party, Collection[Party]] = None,
        act_as: Union[None, Party, Collection[Party]] = None,
        admin: Optional[bool] = False,
        ledger_id: Optional[str] = None,
        application_name: Optional[str] = None,
        oauth_token: Optional[str] = None,
        oauth_token_file: Optional[str] = None,
        ca: Optional[bytes] = None,
        ca_file: Optional[PathLike] = None,
        cert: Optional[bytes] = None,
        cert_file: Optional[PathLike] = None,
        cert_key: Optional[bytes] = None,
        cert_key_file: Optional[PathLike] = None,
        connect_timeout: Optional[TimeDeltaLike] = None,
        use_http_proxy: bool = True,
        logger: Optional[Logger] = None,
        logger_name: Optional[str] = None,
        log_level: Optional[str] = None,
    ) -> "Config":
        """
        Create a :class:`Config` object from the supplied parameters.

        The remote can be configured either by supplying a URL or by supplying a host (and optional
        port and scheme). If none of ``url``, ``host``, ``port``, or ``scheme`` are supplied, then
        environment variables are consulted; if no environment variables are specified either, then
        default values are used.

        The remote can be configured either by supplying a URL or by supplying a host (and optional
        port and scheme). If none of ``url``, ``host``, ``port``, or ``scheme`` are supplied, then
        environment variables are consulted; if no environment variables are specified either, then
        default values are used.

        When the URL scheme is ``http`` or ``https``, dazl will first attempt to connect assuming
        the HTTP JSON API; if this fails, gRPC Ledger API is attempted.

        If ``oauth_token`` is supplied and non-empty, then *token-based access* is used to connect
        to the ledger.

        If other fields are specified, then *property-based access* is used. At least one of
        ``read_as``, ``act_as``, or ``admin`` must be supplied. For the HTTP JSON API,
        ``ledger_id`` MUST be supplied.

        If _no_ fields are specified, this is an error unless environment variables supply an
        alternate source of configuration.

        +-------------------+----------------------+-----------------------------------------------+
        | function argument | environment variable | default value                                 |
        +===================+======================+===============================================+
        | ``url``           | DAML_LEDGER_URL      | localhost:6865                                |
        +-------------------+----------------------+-----------------------------------------------+
        | ``host``          | DAML_LEDGER_HOST     | localhost                                     |
        +-------------------+----------------------+-----------------------------------------------+
        | ``port``          | DAML_LEDGER_PORT     | 6865, unless ``scheme`` is specified:         |
        |                   |                      | * 80 for ``http``                             |
        |                   |                      | * 443 for ``https``                           |
        |                   |                      | * 6865 for ``grpc``                           |
        +-------------------+----------------------+-----------------------------------------------+
        | ``scheme``        | DAML_LEDGER_SCHEME   | | ``https`` for port 443 or 8443              |
        |                   |                      | | ``http`` for port 80, 7575 or 8080          |
        |                   |                      | | ``grpc`` for port 6865                      |
        |                   |                      | | ``https`` for all other ports               |
        +-------------------+----------------------+-----------------------------------------------+

        HTTP(s) proxies (gRPC Ledger API only)
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

        When connecting to the gRPC Ledger API, note that
        `gRPC environment variables
        <https://grpc.github.io/grpc/cpp/md_doc_environment_variables.html>`_
        are always also respected; you can set ``https_proxy``/`http_proxy`` (note the lowercase
        environment variable name). dazl, by default, will _also_ disable usage of a proxy server
        for a ``localhost`` host or a ``127.0.0.1`` host; this can be overridden by passing a value
        of ``use_http_proxy`` to ``True``.

        :param url:
            The URL to connect to. Can be used as an alternative to supplying ``host``, ``port``,
            and ``scheme`` as individual values. Can alternatively be supplied via the environment
            variable ``DAML_LEDGER_URL``.
        :param host:
            The host to connect to. Can be used as an alternative to supplying ``url`` as a combined
            value. Can alternatively be supplied via the environment variable ``DAML_LEDGER_HOST``.
        :param port:
            The port to connect to. Can be used as an alternative to supplying ``url`` as a combined
            value. Can alternatively be supplied via the environment variable ``DAML_LEDGER_PORT``.
        :param scheme:
            The scheme to connect to. Can be used as an alternative to supplying ``url`` as a
            combined value. Can alternatively be supplied via the environment variable
            ``DAML_LEDGER_SCHEME``.
        :param connect_timeout:
            Length of time to wait before giving up connecting to the remote and declaring an error.
            The default value is 30 seconds.
        :param use_http_proxy:
            ``True`` to use an HTTP(S) proxy server if configured; ``False`` to avoid using any
            configured server. If unspecified and the host is ``localhost``, the proxy server is
            avoided; otherwise the proxy server is used.
        :param ca:
            A certificate authority to use to validate the server's certificate. If not supplied,
            the operating system's default trust store is used. Cannot be specified with
            ``ca_file``.
        :param ca_file:
            A file containing the certificate authority to use to validate the server's certificate.
            If not supplied, the operating system's default trust store is used. Cannot be specified
            with ``ca``.
        :param cert:
            A client-side certificate to be used when connecting to a server that requires mutual
            TLS. Cannot be specified with ``cert_file``.
        :param cert_file:
            A file containing the client-side certificate to be used when connecting to a server
            that requires mutual TLS. Cannot be specified with ``cert``.
        :param cert_key:
            A client-side private key to be used when connecting to a server that requires mutual
            TLS. Cannot be specified with ``cert_key_file``.
        :param cert_key_file:
            A client-side private key to be used when connecting to a server that requires mutual
            TLS. Cannot be specified with ``cert_key``.
        :param read_as:
            A party or set of parties on whose behalf (in addition to all parties listed in
            ``act_as``) contracts can be retrieved. Cannot be specified if ``oauth_token`` or
            ``oauth_token_file`` is specified.
        :param act_as:
            A party or set of parties on whose behalf commands should be executed. Parties here are
            also implicitly granted ``read_as`` access as well. Cannot be specified if
            ``oauth_token`` or ``oauth_token_file`` is specified.
        :param admin:
            HTTP JSON API only: allow admin endpoints to be used. This flag is ignored when
            connecting to gRPC Ledger API implementations. Cannot be specified if ``oauth_token`` or
            ``oauth_token_file`` is specified.
        :param ledger_id:
            The ledger ID to connect to. For the HTTP JSON API, this value is required. For the gRPC
            Ledger API, if this value is _not_ supplied, its value will be retrieved from the
            server. Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is specified.
        :param application_name:
            A string that identifies this application. This is used for tracing purposes on the
            server-side. Cannot be specified if ``oauth_token`` or ``oauth_token_file`` is
            specified.
        :param oauth_token:
            The OAuth bearer token to be used on all requests. If specified, no other access
            parameters can be specified.
        :param oauth_token_file:
            A file that contains the OAuth bearer token to be used on all requests. If specified, no
            other access parameters can be specified.
        :param logger:
            The logger to use for connections created from this configuration. If not supplied, a
            logger will be created.
        :param logger_name:
            The name of the logger. Only used if ``logger`` is not provided.
        :param log_level:
            The logging level for the logger. The default is ``warn``. Only used if ``logger`` is
            not provided.
        """
        if logger is None:
            if not logger_name:
                logger_name = f"dazl.conn.{next(id_generator)}"
            logger = logging.getLogger(logger_name)
            if log_level is not None:
                logger.setLevel(log_level)

        url_config = create_url(
            url=url,
            host=host,
            port=port,
            scheme=scheme,
            connect_timeout=connect_timeout,
            use_http_proxy=use_http_proxy,
            logger=logger,
        )

        access_config = create_access(
            read_as=read_as,
            act_as=act_as,
            admin=admin,
            ledger_id=ledger_id,
            application_name=application_name,
            oauth_token=oauth_token,
            oauth_token_file=oauth_token_file,
            logger=logger,
        )

        ssl_config = SSLConfig(
            ca=ca,
            ca_file=ca_file,
            cert=cert,
            cert_file=cert_file,
            cert_key=cert_key,
            cert_key_file=cert_key_file,
            logger=logger,
        )

        return cls(access_config, ssl_config, url_config, logger)

    def __init__(self, access: AccessConfig, ssl: SSLConfig, url: URLConfig, logger: Logger):
        """
        Initialize an instance of :class:`Config`.
        """
        self.access = access
        self.ssl = ssl
        self.url = url
        self.logger = logger



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from ..config import Config
from .conn_aio import Connection

__all__ = ["connect", "Connection"]


def connect(**kwargs):
    """
    Connect to a gRPC Ledger API implementation and return a connection that uses asyncio.
    """
    config = Config.create(**kwargs)
    return Connection(config)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Dict as __Map,
    List as __List,
    Literal as __Literal,
    Mapping as __Mapping,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    overload,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.message import Message as __Message
from google.protobuf.timestamp_pb2 import Timestamp

from .event_pb2 import CreatedEvent, Event, ExercisedEvent
from .trace_context_pb2 import TraceContext

__all__ = [
    "TransactionTree",
    "TreeEvent",
    "Transaction",
]

class TransactionTree(__Message):
    transaction_id: str
    command_id: str
    workflow_id: str
    effective_at: Timestamp
    offset: str
    events_by_id: __Map[str, TreeEvent]
    root_event_ids: __List[str]
    trace_context: TraceContext
    def __init__(self, *, transaction_id: __Optional[str] = ..., command_id: __Optional[str] = ..., workflow_id: __Optional[str] = ..., effective_at: __Optional[Timestamp] = ..., offset: __Optional[str] = ..., events_by_id: __Optional[__Mapping[str, TreeEvent]] = ..., root_event_ids: __Optional[__Sequence[str]] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: TransactionTree) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["transaction_id", "command_id", "workflow_id", "effective_at", "offset", "events_by_id", "root_event_ids", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["transaction_id", "command_id", "workflow_id", "effective_at", "offset", "events_by_id", "root_event_ids", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class TreeEvent(__Message):
    created: CreatedEvent
    exercised: ExercisedEvent
    @overload
    def __init__(self): ...
    @overload
    def __init__(self, *, created: CreatedEvent = ...): ...
    @overload
    def __init__(self, *, exercised: ExercisedEvent = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: TreeEvent) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["kind", "created", "exercised"]) -> bool: ...
    def ClearField(self, field_name: __Literal["kind", "created", "exercised"]) -> None: ...
    def WhichOneof(self, oneof_group: __Literal["kind"]) -> __Literal[None, "created", "exercised"]: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class Transaction(__Message):
    transaction_id: str
    command_id: str
    workflow_id: str
    effective_at: Timestamp
    events: __List[Event]
    offset: str
    trace_context: TraceContext
    def __init__(self, *, transaction_id: __Optional[str] = ..., command_id: __Optional[str] = ..., workflow_id: __Optional[str] = ..., effective_at: __Optional[Timestamp] = ..., events: __Optional[__Sequence[Event]] = ..., offset: __Optional[str] = ..., trace_context: __Optional[TraceContext] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: Transaction) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["transaction_id", "command_id", "workflow_id", "effective_at", "events", "offset", "trace_context"]) -> bool: ...
    def ClearField(self, field_name: __Literal["transaction_id", "command_id", "workflow_id", "effective_at", "events", "offset", "trace_context"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


__all__ = ["DazlError", "DazlWarning", "DazlImportError"]

import warnings


class DazlError(Exception):
    """
    Superclass of errors raised by dazl.
    """


class DazlWarning(Warning):
    """
    Superclass of warnings raised by dazl.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn(
            "DazlWarning is deprecated; it and UnknownTemplateWarning wil be removed in dazl v9",
            DeprecationWarning,
            stacklevel=2,
        )


class DazlImportError(ImportError, DazlError):
    """
    Import error raised when an optional dependency could not be found.
    """

    def __init__(self, missing_module, message):
        super().__init__(message)
        warnings.warn(
            "DazlImportError will be removed in dazl v9; prefer to catch ImportError instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self.missing_module = missing_module



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module prints the metadata obtained from a remote server.
"""

from argparse import ArgumentParser
from typing import Collection

from .. import LOG, Network
from ..client.config import AnonymousNetworkConfig, configure_parser
from ..damlast import DarFile
from ..damlast.lookup import MultiPackageLookup
from ..damlast.protocols import SymbolLookup
from ..pretty import PrettyOptions, get_pretty_printer
from ..protocols.errors import ConnectionTimeoutError, UserTerminateRequest
from ._base import CliCommand

__all__ = ["PrintMetadataCommand"]


class PrintMetadataCommand(CliCommand):
    name = "metadata"

    def parser(self) -> ArgumentParser:
        arg_parser = ArgumentParser()

        configure_parser(arg_parser, parties=False)
        arg_parser.add_argument("--file", help="path to a DAR file", action="append")
        arg_parser.add_argument("--show-hidden", help="show hidden types", action="store_true")
        arg_parser.add_argument("--format", help='one of "daml" or "python"', default="daml")
        return arg_parser

    def execute(self, args) -> int:
        LOG.debug("Executing a metadata fetch...")

        options = PrettyOptions(
            column_width=80, show_hidden_types=args.show_hidden, format=args.format
        )

        if args.file:
            return self.execute_static_metadata(args.file, options)

        config = AnonymousNetworkConfig.get_config(args)

        return self.execute_runtime_metadata(config, options)

    @staticmethod
    def execute_static_metadata(files: Collection[str], options: PrettyOptions) -> int:
        lookup = MultiPackageLookup()
        for file in files:
            with DarFile(file) as dar_file:
                lookup.add_archive(*dar_file.archives())
        _process_metadata(lookup, options)
        return 0

    @staticmethod
    def execute_runtime_metadata(config: "AnonymousNetworkConfig", options: PrettyOptions) -> int:
        try:
            network = Network()
            network.set_config(config)
            network.run_until_complete(_main(network, options))
            return 0
        except UserTerminateRequest:
            return 1
        except ConnectionTimeoutError as ex:
            print(str(ex))
            return 1


async def _main(network: Network, options):
    await network.aio_global().metadata()
    _process_metadata(network.lookup, options)


def _process_metadata(lookup: SymbolLookup, options: PrettyOptions) -> None:
    import sys

    if sys.stdout.isatty():
        try:
            import pygments
            from pygments.formatters.terminal256 import Terminal256Formatter

            formatter = Terminal256Formatter()
        except ImportError:
            pygments = None
            formatter = None
    else:
        pygments = None
        formatter = None

    pretty_printer = get_pretty_printer(options.format, options, lookup)
    if pretty_printer is None:
        raise RuntimeError(f"unknown format: {options.format}")

    code = pretty_printer.render_store()
    lexer = pretty_printer.lexer()

    if pygments is not None and lexer is not None and formatter is not None:
        print(pygments.highlight(code, lexer, formatter))
    else:
        print(code)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import Future, ensure_future, gather, sleep, wait_for
from collections import defaultdict
from dataclasses import asdict, fields
from datetime import timedelta
import logging
from threading import RLock, Thread, current_thread
from typing import (
    Any,
    Awaitable,
    Callable,
    Collection,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from .. import LOG
from ..damlast.daml_lf_1 import PackageRef
from ..damlast.lookup import MultiPackageLookup
from ..damlast.pkgfile import get_dar_package_ids
from ..metrics import MetricEvents
from ..prim import Party, TimeDeltaLike, to_timedelta
from ..protocols import LedgerNetwork
from ..protocols.autodetect import AutodetectLedgerNetwork
from ..protocols.events import BaseEvent, InitEvent, ReadyEvent
from ..scheduler import Invoker, RunLevel
from ._base_model import CREATE_IF_MISSING, NONE_IF_MISSING, IfMissingPartyBehavior
from ._conn_settings import connection_settings
from ._party_client_impl import _PartyClientImpl
from .bots import Bot, BotCollection
from .config import AnonymousNetworkConfig, NetworkConfig, URLConfig
from .errors import DazlPartyMissingError
from .ledger import LedgerMetadata

__all__ = ["_NetworkImpl", "_NetworkRunner"]
T = TypeVar("T")


class _NetworkImpl:

    __slots__ = (
        "lookup",
        "invoker",
        "_lock",
        "_main_thread",
        "_party_clients",
        "_global_impls",
        "_party_impls",
        "bots",
        "_config",
        "_pool",
        "_pool_init",
        "_cached_metadata",
        "_metrics",
    )

    def __init__(self, metrics: "Optional[MetricEvents]" = None):
        self.lookup = MultiPackageLookup()
        self.invoker = Invoker()

        self._lock = RLock()
        self._main_thread = None  # type: Optional[Thread]
        self._pool = None  # type: Optional[LedgerNetwork]
        self._pool_init = self.invoker.create_future()
        if metrics is None:
            # create a default set of metrics
            try:
                # noinspection PyUnresolvedReferences
                from ..metrics.prometheus import PrometheusMetricEvents

                self._metrics = PrometheusMetricEvents.default()
            except ImportError:
                self._metrics = MetricEvents()

        # Separately, keep a reference to metadata because it's accessed quite frequently and it is
        # useful for this to be available even when the pool is closed
        self._cached_metadata = None  # type: Optional[LedgerMetadata]

        self._party_clients = defaultdict(dict)  # type: Dict[type, Dict[Party, Any]]
        self._global_impls = dict()  # type: Dict[type, Any]
        self._party_impls = dict()  # type: Dict[Party, _PartyClientImpl]

        self._config = dict()  # type: Dict[str, Any]
        self.bots = BotCollection(None)

    def set_config(self, *configs: "Union[NetworkConfig, AnonymousNetworkConfig]", **kwargs):
        for config in configs:
            self._config.update({k: v for k, v in asdict(config).items() if v is not None})
        self._config.update({k: v for k, v in kwargs.items() if v is not None})
        LOG.debug("Configuration for this network: %r", self._config)

    def get_config_raw(self, key: str, default: Any) -> Any:
        return self._config.get(key, default)

    def resolved_config(self) -> "NetworkConfig":
        config = self.resolved_anonymous_config()
        return NetworkConfig(
            **asdict(config),
            parties=tuple(
                party_impl.resolved_config() for party_impl in self._party_impls.values()
            ),
        )

    def resolved_anonymous_config(self) -> "AnonymousNetworkConfig":
        return AnonymousNetworkConfig.parse_kwargs(**self._config)

    def freeze(self):
        """
        Freeze configuration, and assume the current run loop can be used to schedule dazl
        coroutines. Once this method is called, ``aio_run()`` must be called instead of ``start()``
        in order to run dazl.
        """
        with self._lock:
            self.invoker.set_context_as_current()
            config = self.resolved_config()
            LOG.debug("Resolved config: %s", config)

        # From this point on, we're assuming we're on an asyncio event loop so locking is no longer
        # required
        if self._pool is None:
            from ..protocols import LedgerConnectionOptions

            # If the connect timeout is non-positive, assume the user intended for there to not be
            # a timeout at all
            connect_timeout = to_timedelta(config.connect_timeout)
            if connect_timeout <= timedelta(0):
                connect_timeout = None

            package_lookup_timeout = to_timedelta(config.package_lookup_timeout)
            if package_lookup_timeout <= timedelta(0):
                package_lookup_timeout = None

            options = LedgerConnectionOptions(
                lookup=self.lookup,
                connect_timeout=connect_timeout,
                package_lookup_timeout=package_lookup_timeout,
                eager_package_fetch=config.eager_package_fetch,
            )

            self._pool = pool = AutodetectLedgerNetwork(self.invoker, options)
            self._pool_init.set_result(pool)

        return config

    async def aio_run(self, *coroutines) -> None:
        """
        Coroutine where all network activity is scheduled from.

        :param coroutines:
            Optional additional coroutines to run. ``aio_run`` is only considered complete once all
            of the additional coroutines are also finished.
        """
        from ..metrics.instrumenters import AioLoopPerfMonitor

        config = self.freeze()
        if self._pool is None:
            raise RuntimeError("freeze() did not kick off a network pool!")

        site = None
        with AioLoopPerfMonitor(self._metrics.loop_responsiveness):
            if config.server_port is not None:
                LOG.info("Opening port %s for metrics...", config.server_port)
                try:
                    from aiohttp import web

                    from ..server import get_app

                    app = get_app(self)
                    app_runner = web.AppRunner(app)
                    await app_runner.setup()
                    site = web.TCPSite(app_runner, config.server_host, config.server_port)
                    ensure_future(site.start())
                    LOG.info(
                        "Listening on %s:%s for metrics.", config.server_host, config.server_port
                    )
                except ImportError:
                    LOG.warning("Could not start metrics because aiohttp is not installed")
            else:
                LOG.info(
                    "No server_port configuration was specified, so metrics and other stats "
                    "will not be served."
                )

            try:
                runner = _NetworkRunner(self._pool, self, coroutines)
                await runner.run()
            finally:
                await self._pool.close()
                if site is not None:
                    await site.stop()
                await self.invoker.shutdown(0)
                self._pool = None

    def start(self, daemon: bool) -> None:
        """
        Create a background thread, start an event loop, and run the application.
        """
        from asyncio import new_event_loop, set_event_loop

        def background_main():
            LOG.info("Starting an event loop on a background thread %r...", current_thread().name)

            try:
                loop = new_event_loop()
                set_event_loop(loop)

                loop.run_until_complete(self.aio_run())
            except:
                LOG.exception("The main event loop died!")

            LOG.info("The main event loop has finished.")

        with self._lock:
            if self._main_thread is not None:
                raise RuntimeError("start() called more than once")
            self._main_thread = Thread(
                target=background_main, daemon=daemon, name=f"dazl:main-{id(self):016x}"
            )

        self._main_thread.start()

    def shutdown(self) -> None:
        """
        Indicate that the client and its connected parties should be stopped as soon as is cleanly
        possible.

        This method can safely be called from any thread, but it must be called after start().
        """
        with self._lock:
            LOG.info("Shutting down...")
            self.invoker.handle_sigint()

    def abort(self) -> None:
        """
        Indicate that the client and its connected parties should be stopped as immediately.

        This method can safely be called from any thread, but it must be called after start().
        """
        with self._lock:
            LOG.info("Aborting...")
            self.invoker.handle_sigquit()

    def join(self, timeout=None):
        """

        :param timeout:
        :return:
        """
        with self._lock:
            thread = self._main_thread

        if thread is not None:
            thread.join(timeout=timeout)

    def global_impl(self, ctor: "Callable[[_NetworkImpl], T]") -> T:
        with self._lock:
            inst = self._global_impls.get(ctor)  # type: ignore
            if inst is None:
                inst = ctor(self)
                self._global_impls[ctor] = inst  # type: ignore
            return inst

    def party_impl_wrapper(self, party: Party, ctor: "Callable[[_PartyClientImpl], T]") -> "T":
        return self.party_impl(party, ctor=ctor)

    def party_impl(self, party, ctor=None, if_missing: IfMissingPartyBehavior = CREATE_IF_MISSING):
        """
        Return a :class:`_PartyClientImpl` object for the specified party, possibly wrapping it in a
        more high-level friendly API.

        :param party:
            The party to get (or create) a connection for.
        :param ctor:
            The constructor function to use to create a wrapper, or ``None`` to return a raw
            :class:`_PartyClientImpl`.
        :param if_missing:
            The behavior of this method in the case that an entry for the Party does not yet exist:
            ``CREATE_IF_MISSING``: if the party does not yet exist, create an implementation for it.
            ``NONE_IF_MISSING``: if the party does not yet exist, return ``None``.
            ``EXCEPTION_IF_MISSING``: if the party does not yet exist, throw an exception.
        :return:
            Either a :class:`_PartyClientImpl`, the specified wrapper type, or ``None``.
        """
        with self._lock:
            impl = self._party_impls.get(party)
            if impl is None:
                if if_missing == CREATE_IF_MISSING:
                    impl = _PartyClientImpl(self, party)
                    self._party_impls[party] = impl
                elif if_missing == NONE_IF_MISSING:
                    return None
                else:
                    raise DazlPartyMissingError(party)

            if ctor is None:
                return impl

            mm = self._party_clients[ctor]
            inst = mm.get(party)
            if inst is None:
                inst = ctor(impl)
                mm[party] = inst
            return inst

    def parties(self) -> "Collection[Party]":
        """
        Return a snapshot of the set of parties that exist right now.
        """
        with self._lock:
            return list(self._party_impls)

    def party_impls(self) -> "Collection[_PartyClientImpl]":
        with self._lock:
            return list(self._party_impls.values())

    def find_bot(self, bot_id) -> "Optional[Bot]":
        """
        Return the bot of the specified ID, or ``None`` if no matching bot is found.
        """
        for party_impl in self.party_impls():
            for bot in party_impl.bots:
                if bot.id == bot_id:
                    return bot
        return None

    def simple_metadata(self, timeout: "TimeDeltaLike") -> "LedgerMetadata":
        """
        Retrieve metadata about the ledger.

        If you are using the ``asyncio``-flavored version of the API, you _must_ use
        :meth:`aio_metadata` instead.
        """
        if self._cached_metadata is not None:
            return self._cached_metadata

        if self._pool is None:
            raise ValueError(
                "simple_metadata must be called only after the global connection has started"
            )

        with self._lock:
            return self.invoker.run_in_loop(self._pool.ledger, timeout=timeout)

    async def aio_metadata(self) -> "LedgerMetadata":
        """
        Coroutine version of :meth:`metadata`.
        """
        pool = await self._pool_init
        return await pool.ledger()

    async def upload_package(self, contents: bytes, timeout: "TimeDeltaLike") -> None:
        """
        Ensure packages specified by the given byte array are loaded on the remote server. This
        method only returns once packages are reported by the Package Service.

        :param contents: Contents to attempt to upload.
        :param timeout: Length of time before giving up.
        """
        pool = await self._pool_init
        await self.connect_anonymous()
        package_ids = await self.invoker.run_in_executor(lambda: get_dar_package_ids(contents))
        await pool.upload_package(contents)
        await self.ensure_package_ids(package_ids, timeout)

    async def ensure_package_ids(
        self, package_ids: "Collection[PackageRef]", timeout: "TimeDeltaLike"
    ):
        await wait_for(
            self.__ensure_package_ids(package_ids), to_timedelta(timeout).total_seconds()
        )

    async def __ensure_package_ids(self, package_ids: "Collection[PackageRef]"):
        metadata = await self.aio_metadata()
        await gather(*(metadata.package_loader.load(pkg_ref) for pkg_ref in package_ids))

    # region Event Handler Management

    # noinspection PyShadowingBuiltins

    def add_event_handler(self, key, handler: "Callable[[BaseEvent], None]") -> "Bot":
        """
        Add an event handler to a specific event. Unlike event listeners on party clients, these
        event handlers are not allowed to return anything in response to handling an event.
        """
        return self.bots.add_single(key, handler, None)

    def emit_event(self, data: BaseEvent) -> "Awaitable[Any]":
        return self.bots.notify(data)

    # endregion

    async def connect_anonymous(self):
        config_names = {f.name for f in fields(URLConfig)}
        config = URLConfig(**{k: v for k, v in self._config.items() if k in config_names})
        settings, url_prefix = connection_settings(
            config.url,
            None,
            verify_ssl=config.verify_ssl,
            ca_file=config.ca_file,
            cert_file=config.cert_file,
            cert_key_file=config.cert_key_file,
            enable_http_proxy=config.enable_http_proxy,
        )
        await self._pool.connect_anonymous(settings, url_prefix)


class _NetworkRunner:
    def __init__(self, pool: LedgerNetwork, network_impl: "_NetworkImpl", user_coroutines):
        self.pool = pool
        self.initialized_parties = set()  # type: Set[Party]
        self._network_impl = network_impl
        self._config = self._network_impl.resolved_config()
        self._read_completions = []  # type: List[Future]
        self._user_coroutines = [ensure_future(coro) for coro in user_coroutines]
        self._bot_coroutines = []  # type: List[Future]
        self._writers = []  # type: List[Future]

    async def run(self):
        LOG.debug("Running a Network with config: %r", self._config)
        prev_offset = None
        keep_running = True

        self._bot_coroutines.append(ensure_future(self._network_impl.bots._main()))

        while keep_running:
            party_impls = {pi.party: pi for pi in self._network_impl.party_impls()}
            if not party_impls:
                await self._network_impl.connect_anonymous()

            # make sure that all parties are initialized; we do this on EVERY tick because new
            # parties might come later and they'll need to be initialized too
            uninitialized_parties = set(party_impls) - self.initialized_parties
            # TODO: Also handle removed parties at some point
            if uninitialized_parties:
                await self._init(
                    [party_impls[party] for party in uninitialized_parties], prev_offset
                )
            self.initialized_parties |= uninitialized_parties

            current_offset, keep_running = await self.beat(list(party_impls.values()))
            if keep_running:
                if prev_offset == current_offset:
                    await sleep(0.5)
                else:
                    prev_offset = current_offset

        # prohibit any more command submissions
        LOG.debug("network_run: stopping bots gracefully...")
        for pi in self._network_impl.party_impls():
            pi.bots.stop_all()
        self._network_impl.bots.stop_all()

        LOG.debug("network_run: stopping its writers gracefully...")
        # TODO: This needs to be coordinated better among the list of writers we ACTUALLY started
        for pi in self._network_impl.party_impls():
            pi.stop_writer()

        # wait for writers to finish their writing
        await gather(*self._writers, return_exceptions=True)
        LOG.info("network_run: finished.")
        self._writers = []

    async def beat(self, party_impls: "Collection[_PartyClientImpl]") -> Tuple[Optional[str], bool]:
        """
        Called periodically to schedule reads and writes.
        """
        from ._reader_sync import run_iteration

        offset, completions = await run_iteration(party_impls)

        self._read_completions = [
            fut for fut in (*self._read_completions, *completions) if not fut.done()
        ]
        self._user_coroutines = [fut for fut in self._user_coroutines if not fut.done()]

        # If there are no more commands in flight, there is no more activity
        if self._network_impl.invoker.level >= RunLevel.TERMINATE_GRACEFULLY:
            LOG.info("network_run terminating on user request...")
            return offset, False
        elif self._network_impl.invoker.level >= RunLevel.RUN_UNTIL_IDLE:
            if not self._read_completions and not self._user_coroutines:
                if all(pi.writer_idle() for pi in party_impls):
                    LOG.info(
                        "network_run: terminating because all writers are idle and all "
                        "clients have caught up to offset %s",
                        offset,
                    )
                    return offset, False

        return offset, True

    async def _init(
        self, party_impls: "Collection[_PartyClientImpl]", first_offset: "Optional[str]"
    ) -> Optional[str]:
        # region log
        if LOG.isEnabledFor(logging.DEBUG):
            parties = [pi.party for pi in party_impls]
            if first_offset is None:
                LOG.debug("Starting first-time initialization for parties: %s", parties)
            else:
                LOG.debug(
                    "Starting initialization for parties: %s, advancing to offset: %s",
                    parties,
                    first_offset,
                )
        # endregion

        # establish ledger connections
        if party_impls:
            # noinspection PyProtectedMember
            self._bot_coroutines.extend(
                ensure_future(party_impl.bots._main()) for party_impl in party_impls
            )
            await gather(*(party_impl.connect_in(self.pool) for party_impl in party_impls))
        elif first_offset is None:
            # establish a single null connection that will be used only to ensure that metadata can be
            # fetched
            party_impl = _PartyClientImpl(self._network_impl, Party("\x00"))
            await party_impl.connect_in(self.pool)
        else:
            # trivially, there is nothing to do for init because there are no parties to initialize
            # and this isn't our first rodeo
            return first_offset

        # get metadata about the ledger
        metadata = await self.pool.ledger()
        self._network_impl._cached_metadata = metadata

        # start writers early to potentially service commands off the back of these events
        self._writers.extend(ensure_future(party_impl.main_writer()) for party_impl in party_impls)

        # Raise the 'init' event.
        init_futs = []
        if first_offset is None:
            init_evt = InitEvent(
                None,
                None,
                None,
                metadata.ledger_id,
                self._network_impl.lookup,
                metadata._store,
            )
            init_futs.append(ensure_future(self._network_impl.emit_event(init_evt)))
        for party_impl in party_impls:
            init_futs.append(ensure_future(party_impl.initialize(None, metadata)))

        # TODO: Consider what should join on init_futs; we can't join here because otherwise this
        #  blocks the readers from reading transactions that would cause these events to fully
        #  resolve

        from ._reader_sync import read_initial_acs

        offset = await read_initial_acs(party_impls)

        LOG.debug('Preparing to raise the "ready" event...')
        # Raise the 'ready' event.
        ready_futs = []
        if first_offset is None:
            ready_evt = ReadyEvent(
                None,
                None,
                None,
                metadata.ledger_id,
                self._network_impl.lookup,
                metadata._store,
                offset,
            )
            ready_futs.append(ensure_future(self._network_impl.emit_event(ready_evt)))
        for party_impl in party_impls:
            ready_futs.append(ensure_future(party_impl.emit_ready(metadata, None, offset)))
        for party_impl in party_impls:
            party_impl.ready().set_result(None)

        # TODO: Consider what should join on ready_futs; we can't join here because otherwise this
        #  blocks the readers from reading transactions that would cause these events to fully
        #  resolve

        # Return the metadata offset for later
        LOG.debug("Raised the ready event, and caught up to %s.", offset)
        return offset



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..client._conn_settings import OAuthSettings


__all__ = ["oauth_flow"]


async def oauth_flow(settings: "OAuthSettings") -> "OAuthSettings":
    """
    Implementation of an OAuth flow that ensures that a token is provided.

    :param settings:
    :return:
    """
    import requests

    from ..client._conn_settings import OAuthSettings

    if not settings.token:

        if settings.auth_audience == None:
            logging.error("ERROR: Need to supply an oAuth audience")
            raise ValueError("ERROR: Need to supply an oAuth audience")

        headers = {"Accept": "application/json"}
        data = {
            "client_id": settings.client_id,
            "client_secret": settings.client_secret,
            "audience": settings.auth_audience,
            "grant_type": "client_credentials",
        }

        if settings.token_uri is None:
            raise ValueError("missing a token URI")
        response = None
        try:
            response = requests.post(
                settings.token_uri,
                headers=headers,
                data=data,
                auth=None,
                verify=settings.auth_ca_file,
            )
        except Exception as ex:
            logging.info(ex)
            raise ValueError("Unable to get token at this time")

        if response.status_code != 200:
            logging.error("ERROR: Unable to retrieve token. Exiting")
            raise ValueError("Unable to get token from oAuth source")

        json = response.json()
        return_settings = OAuthSettings(
            client_id=settings.client_id,
            client_secret=settings.client_secret,
            token=json["access_token"],
            token_uri=settings.token_uri,
            refresh_token=settings.refresh_token,
            id_token=settings.id_token,
            auth_url=settings.auth_url,
            redirect_uri=settings.redirect_uri,
            auth_audience=settings.auth_audience,
            auth_ca_file=settings.auth_ca_file,
        )

        return return_settings

    else:
        return settings



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
This module contains the mapping between gRPC calls and Python/dazl types.
"""

import asyncio
from typing import AbstractSet, Any, AsyncIterable, Collection, Mapping, Optional, Sequence, Union
import uuid
import warnings

from grpc.aio import Channel, UnaryStreamCall

from ..._gen.com.daml.ledger.api.v1.active_contracts_service_pb2 import (
    GetActiveContractsRequest as G_GetActiveContractsRequest,
)
from ..._gen.com.daml.ledger.api.v1.active_contracts_service_pb2_grpc import (
    ActiveContractsServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.admin.package_management_service_pb2 import (
    UploadDarFileRequest as G_UploadDarFileRequest,
)
from ..._gen.com.daml.ledger.api.v1.admin.package_management_service_pb2_grpc import (
    PackageManagementServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.admin.party_management_service_pb2 import (
    AllocatePartyRequest as G_AllocatePartyRequest,
)
from ..._gen.com.daml.ledger.api.v1.admin.party_management_service_pb2_grpc import (
    PartyManagementServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.command_service_pb2 import (
    SubmitAndWaitRequest as G_SubmitAndWaitRequest,
)
from ..._gen.com.daml.ledger.api.v1.command_service_pb2_grpc import CommandServiceStub
from ..._gen.com.daml.ledger.api.v1.commands_pb2 import Command as G_Command, Commands as G_Commands
from ..._gen.com.daml.ledger.api.v1.ledger_identity_service_pb2 import (
    GetLedgerIdentityRequest as G_GetLedgerIdentityRequest,
)
from ..._gen.com.daml.ledger.api.v1.ledger_identity_service_pb2_grpc import (
    LedgerIdentityServiceStub,
)
from ..._gen.com.daml.ledger.api.v1.package_service_pb2 import (
    GetPackageRequest as G_GetPackageRequest,
    ListPackagesRequest as G_ListPackagesRequest,
)
from ..._gen.com.daml.ledger.api.v1.package_service_pb2_grpc import PackageServiceStub
from ..._gen.com.daml.ledger.api.v1.transaction_filter_pb2 import (
    TransactionFilter as G_TransactionFilter,
)
from ..._gen.com.daml.ledger.api.v1.transaction_service_pb2 import (
    GetTransactionsRequest as G_GetTransactionsRequest,
)
from ..._gen.com.daml.ledger.api.v1.transaction_service_pb2_grpc import TransactionServiceStub
from ...damlast.daml_lf_1 import PackageRef, TypeConName
from ...prim import LEDGER_STRING_REGEX, ContractData, ContractId, Party
from ...query import Query
from .._offsets import UNTIL_END, LedgerOffsetRange, from_offset_until_forever
from ..api_types import ArchiveEvent, Boundary, Command, CreateEvent, ExerciseResponse, PartyInfo
from ..config import Config
from ..config.access import PropertyBasedAccessConfig
from ..errors import ProtocolWarning
from ..stream_aio import QueryStreamBase
from .channel import create_channel
from .codec_aio import Codec

__all__ = ["Connection", "QueryStream"]


class Connection:
    def __init__(self, config: Config):
        self._config = config
        self._logger = config.logger
        self._channel = create_channel(config)
        self._codec = Codec(self)

    @property
    def config(self) -> Config:
        return self._config

    @property
    def channel(self) -> Channel:
        """
        Provides access to the underlying gRPC channel.
        """
        return self._channel

    @property
    def codec(self) -> Codec:
        return self._codec

    async def __aenter__(self) -> "Connection":
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    async def open(self) -> None:
        """
        Does final validation of the token, including possibly fetching the ledger ID if it is not
        yet known.
        """
        if not self._config.access.ledger_id:
            # most calls require a ledger ID; if it wasn't supplied as part of our token or we were
            # never given a token in the first place, fetch the ledger ID from the destination
            stub = LedgerIdentityServiceStub(self._channel)
            response = await stub.GetLedgerIdentity(G_GetLedgerIdentityRequest())
            if isinstance(self._config.access, PropertyBasedAccessConfig):
                self._logger.info("Connected to gRPC Ledger API, ledger ID: %s", response.ledger_id)
                self._config.access.ledger_id = response.ledger_id
            else:
                raise ValueError("when using token-based access, the token must contain ledger ID")

    async def close(self) -> None:
        """
        Close the underlying channel. Once the channel is closed, future command submissions,
        streams in progress, and any future streams will fail.
        """
        await self._channel.close()

    # region Write API

    async def do_commands(
        self,
        commands: Union[Command, Sequence[Command]],
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> "None":
        """
        Submit one or more commands to the Ledger API.

        You should generally prefer trying to use :meth:`create`, :meth:`exercise`,
        :meth:`exercise_by_key`, or :meth:`create_and_exercise`, as they are available over both
        the gRPC Ledger API and HTTP JSON API; additionally those methods can provide more
        information about what happened.

        This method can be used to submit multiple disparate commands as a single transaction, but
        if you find yourself needing to do this, you may want to consider moving more of your logic
        into Daml so that only a single command is needed from the outside in order to satisfy your
        use case.
        """
        if commands is None:
            return
        elif isinstance(commands, Command):
            commands = [commands]

        stub = CommandServiceStub(self.channel)

        commands_pb = await asyncio.gather(*map(self._codec.encode_command, commands))
        request = G_SubmitAndWaitRequest(
            commands=G_Commands(
                ledger_id=self._config.access.ledger_id,
                application_id=self._config.access.application_name,
                command_id=self._command_id(command_id),
                workflow_id=self._workflow_id(workflow_id),
                party=self._ensure_act_as(),
                commands=commands_pb,
                act_as=self._config.access.act_as,
                read_as=self._config.access.read_only_as,
            )
        )
        await stub.SubmitAndWait(request)

    async def create(
        self,
        template_id: Union[str, TypeConName],
        payload: ContractData,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> CreateEvent:
        """
        Create a contract for a given template.

        :param template_id: The template of the contract to be created.
        :param payload: Template arguments for the contract to be created.
        :param workflow_id: An optional workflow ID.
        :param command_id: An optional command ID. If unspecified, a random one will be created.
        """
        stub = CommandServiceStub(self.channel)

        commands = [G_Command(create=await self._codec.encode_create_command(template_id, payload))]
        request = G_SubmitAndWaitRequest(
            commands=G_Commands(
                ledger_id=self._config.access.ledger_id,
                application_id=self._config.access.application_name,
                command_id=self._command_id(command_id),
                workflow_id=self._workflow_id(workflow_id),
                party=self._ensure_act_as(),
                commands=commands,
                act_as=self._config.access.act_as,
                read_as=self._config.access.read_only_as,
            )
        )
        response = await stub.SubmitAndWaitForTransaction(request)

        return await self._codec.decode_created_event(response.transaction.events[0].created)

    async def exercise(
        self,
        contract_id: ContractId,
        choice_name: str,
        argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> ExerciseResponse:
        """
        Exercise a choice on a contract identified by its contract ID.

        :param contract_id: The contract ID of the contract to exercise.
        :param choice_name: The name of the choice to exercise.
        :param argument: The choice arguments. Can be omitted for choices that take no argument.
        :param workflow_id: An optional workflow ID.
        :param command_id: An optional command ID. If unspecified, a random one will be created.
        :return: A response
        """
        stub = CommandServiceStub(self.channel)

        commands = [
            G_Command(
                exercise=await self._codec.encode_exercise_command(
                    contract_id, choice_name, argument
                )
            )
        ]
        request = self._submit_and_wait_request(commands, workflow_id, command_id)
        response = await stub.SubmitAndWaitForTransactionTree(request)

        return await self._codec.decode_exercise_response(response.transaction)

    async def create_and_exercise(
        self,
        template_id: Union[str, TypeConName],
        payload: ContractData,
        choice_name: str,
        argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> ExerciseResponse:
        stub = CommandServiceStub(self.channel)

        commands = [
            G_Command(
                createAndExercise=await self._codec.encode_create_and_exercise_command(
                    template_id, payload, choice_name, argument
                )
            )
        ]
        request = self._submit_and_wait_request(commands, workflow_id, command_id)
        response = await stub.SubmitAndWaitForTransactionTree(request)

        return await self._codec.decode_exercise_response(response.transaction)

    async def exercise_by_key(
        self,
        template_id: Union[str, TypeConName],
        choice_name: str,
        key: Any,
        argument: Optional[ContractData] = None,
        *,
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> "ExerciseResponse":
        stub = CommandServiceStub(self.channel)

        commands = [
            G_Command(
                exerciseByKey=await self._codec.encode_exercise_by_key_command(
                    template_id, choice_name, key, argument
                )
            )
        ]
        request = self._submit_and_wait_request(commands, workflow_id, command_id)
        response = await stub.SubmitAndWaitForTransactionTree(request)

        return await self._codec.decode_exercise_response(response.transaction)

    async def archive(self, contract_id: ContractId) -> ArchiveEvent:
        await self.exercise(contract_id, "Archive")
        return ArchiveEvent(contract_id)

    async def archive_by_key(self, template_id: str, key: Any) -> ArchiveEvent:
        response = await self.exercise_by_key(template_id, "Archive", key)
        return next(iter(event for event in response.events if isinstance(event, ArchiveEvent)))

    def _ensure_act_as(self) -> Party:
        act_as_party = next(iter(self._config.access.act_as), None)
        if not act_as_party:
            raise ValueError("current access rights do not include any act-as parties")
        return act_as_party

    @staticmethod
    def _workflow_id(workflow_id: Optional[str]) -> Optional[str]:
        if workflow_id:
            if not LEDGER_STRING_REGEX.match(workflow_id):
                raise ValueError("workflow_id must be a valid ledger string")
            return workflow_id
        else:
            return None

    @staticmethod
    def _command_id(command_id: Optional[str]) -> str:
        if command_id:
            if not LEDGER_STRING_REGEX.match(command_id):
                raise ValueError("command_id must be a valid ledger string")
            return command_id
        else:
            return uuid.uuid4().hex

    def _submit_and_wait_request(
        self,
        commands: Collection[G_Command],
        workflow_id: Optional[str] = None,
        command_id: Optional[str] = None,
    ) -> G_SubmitAndWaitRequest:
        return G_SubmitAndWaitRequest(
            commands=G_Commands(
                ledger_id=self._config.access.ledger_id,
                application_id=self._config.access.application_name,
                command_id=self._command_id(command_id),
                workflow_id=self._workflow_id(workflow_id),
                party=self._ensure_act_as(),
                commands=commands,
                act_as=self._config.access.act_as,
                read_as=self._config.access.read_only_as,
            )
        )

    # endregion

    # region Read API

    def query(self, __template_id: str = "*", __query: Query = None) -> "QueryStream":
        """
        Return the create events from the active contract set service as a stream.

        If you find yourself repeatedly calling :meth:`query` or :meth:`query_many` over the same
        set of templates, you may want to consider :class:`ACS` instead, which is a utility class
        that helps you maintain a "live" state of the ACS.

        :param __template_id:
            The name of the template for which to fetch contracts.
        :param __query:
            A filter to apply to the set of returned contracts.
        """
        return QueryStream(self, {__template_id: __query}, UNTIL_END)

    def query_many(self, __queries: Optional[Mapping[str, Query]] = None) -> "QueryStream":
        """
        Return the create events from the active contract set service as a stream.

        If you find yourself repeatedly calling :meth:`query` or :meth:`query_many` over the same
        set of templates, you may want to consider :class:`ACS` instead, which is a utility class
        that helps you maintain a "live" state of the ACS.

        :param __queries:
            A map of template IDs to filter to apply to the set of returned contracts.
        """
        return QueryStream(self, __queries, UNTIL_END)

    def stream(
        self, __template_id: str = "*", __query: Query = None, *, offset: Optional[str] = None
    ) -> "QueryStream":
        """
        Stream create/archive events.

        When ``offset`` is ``None``, create events from the active contract set are returned first,
        followed by a continuous stream of updates (creates/archives).

        Otherwise, ``offset`` can be supplied to resume a stream from a prior point where a
        ``Boundary`` was returned from a previous stream.

        :param __template_id:
            The name of the template for which to fetch contracts.
        :param __query:
            A filter to apply to the set of returned contracts. Note that this does not filter
            :class:`ArchiveEvent`; readers of the stream MUST be able to cope with "mismatched"
            archives that come from the result of applying a filter.
        :param offset:
            An optional offset at which to start receiving events. If ``None``, start from the
            beginning.
        """
        return QueryStream(self, {__template_id: __query}, from_offset_until_forever(offset))

    def stream_many(
        self, __queries: Optional[Mapping[str, Query]] = None, *, offset: Optional[str] = None
    ) -> "QueryStream":
        """
        Stream create/archive events from more than one template ID in the same stream.

        When ``offset`` is ``None``, create events from the active contract set are returned first,
        followed by a continuous stream of updates (creates/archives).

        Otherwise, ``offset`` can be supplied to resume a stream from a prior point where a
        ``Boundary`` was returned from a previous stream.

        :param __queries:
            A map of template IDs to filter to apply to the set of returned contracts. Note that
            this does not filter :class:`ArchiveEvent`; readers of the stream MUST be able to cope
            with "mismatched" archives that come from the result of applying a filter.
        :param offset:
            An optional offset at which to start receiving events. If ``None``, start from the
            beginning.
        """
        return QueryStream(self, __queries, from_offset_until_forever(offset))

    # endregion

    # region Party Management calls

    async def allocate_party(
        self, identifier_hint: str = None, display_name: str = None
    ) -> "PartyInfo":
        """
        Allocate a new party.
        """
        stub = PartyManagementServiceStub(self.channel)
        request = G_AllocatePartyRequest(
            party_id_hint=Party(identifier_hint) if identifier_hint else None,
            display_name=display_name,
        )
        response = await stub.AllocateParty(request)
        return Codec.decode_party_info(response.party_details)

    async def list_known_parties(self) -> Sequence[PartyInfo]:
        stub = PartyManagementServiceStub(self.channel)
        response = await stub.ListKnownParties()
        return [Codec.decode_party_info(pd) for pd in response.party_details]

    # endregion

    # region Package Management calls

    async def get_package(self, package_id: PackageRef) -> bytes:
        stub = PackageServiceStub(self.channel)
        request = G_GetPackageRequest(
            ledger_id=self._config.access.ledger_id, package_id=package_id
        )
        response = await stub.GetPackage(request)
        return response.archive_payload

    async def list_package_ids(self) -> AbstractSet[PackageRef]:
        stub = PackageServiceStub(self.channel)
        request = G_ListPackagesRequest(ledger_id=self._config.access.ledger_id)
        response = await stub.ListPackages(request)
        return frozenset({PackageRef(pkg_id) for pkg_id in response.package_ids})

    async def upload_package(self, contents: bytes) -> None:
        stub = PackageManagementServiceStub(self.channel)
        request = G_UploadDarFileRequest(dar_file=contents)
        await stub.UploadDarFile(request)
        return

    # endregion


class QueryStream(QueryStreamBase):
    def __init__(
        self,
        conn: Connection,
        queries: Optional[Mapping[str, Optional[Query]]],
        offset_range: LedgerOffsetRange,
    ):
        self.conn = conn
        self._queries = queries
        self._offset_range = offset_range

        self._filter = None
        self._response_stream = None  # type: Optional[UnaryStreamCall]

    async def close(self) -> None:
        if self._response_stream is not None:
            self._response_stream.cancel()
            self._response_stream = None

    async def items(self):
        """
        Return an asynchronous stream of events.

        .. code-block:: python

            async with conn.query('SampleApp:Iou') as query:
                async for r in query:
                    print(f"Offset: {r.offset}")
                    for event in r.events:
                        print(f"  Event: {event}")

        :return:
            A stream of responses, where each response contains one or more events at a particular
            offset.

            At least one initial :class:`Boundary` is always returned, even if the stream is empty.
            In this case, the first returned object is a :class:`Boundary` with ``offset=None``.
        """
        log = self.conn.config.logger

        filters = await self.conn.codec.encode_filters(self._queries)
        filters_by_party = {party: filters for party in self.conn.config.access.read_as}
        tx_filter_pb = G_TransactionFilter(filters_by_party=filters_by_party)

        try:
            offset = None
            if self._offset_range.begin is None:
                # when starting from the beginning of the ledger, the Active Contract Set service
                # lets us catch up more quickly than having to parse every create/archive event
                # ourselves
                log.debug("Reading from the ACS...")
                async for event in self._acs_events(tx_filter_pb):
                    if isinstance(event, CreateEvent):
                        await self._emit_create(event)
                    elif isinstance(event, Boundary):
                        offset = event.offset
                        await self._emit_boundary(event)
                    else:
                        warnings.warn(f"Received an unknown event: {event}", ProtocolWarning)
                    yield event
            else:
                log.debug(
                    "Skipped reading from the ACS because begin offset is %r",
                    self._offset_range.begin,
                )

            if self._offset_range != UNTIL_END:
                # now start returning events as they come off the transaction stream; note this
                # stream will never naturally close, so it's on the caller to call close() or to
                # otherwise exit our current context
                log.debug("Reading a transaction stream: %s", self._offset_range)
                async for event in self._tx_events(tx_filter_pb, offset):
                    if isinstance(event, CreateEvent):
                        await self._emit_create(event)
                    elif isinstance(event, ArchiveEvent):
                        await self._emit_archive(event)
                    elif isinstance(event, Boundary):
                        await self._emit_boundary(event)
                    else:
                        warnings.warn(f"Received an unknown event: {event}", ProtocolWarning)
                    yield event
            else:
                log.debug(
                    "Not reading from transaction stream because we were only asked for a snapshot."
                )
        finally:
            await self.close()

    async def _acs_events(
        self, filter_pb: G_TransactionFilter
    ) -> AsyncIterable[Union[CreateEvent, Boundary]]:
        stub = ActiveContractsServiceStub(self.conn.channel)

        request = G_GetActiveContractsRequest(
            ledger_id=self.conn.config.access.ledger_id, filter=filter_pb
        )
        self._response_stream = response_stream = stub.GetActiveContracts(request)

        offset = None
        async for response in response_stream:
            for event in response.active_contracts:
                yield await self.conn.codec.decode_created_event(event)
            # for ActiveContractSetResponse messages, only the last offset is actually relevant
            offset = response.offset
        yield Boundary(offset)

    async def _tx_events(
        self, filter_pb: G_TransactionFilter, begin_offset: Optional[str]
    ) -> AsyncIterable[Union[CreateEvent, ArchiveEvent, Boundary]]:
        stub = TransactionServiceStub(self.conn.channel)

        request = G_GetTransactionsRequest(
            ledger_id=self.conn.config.access.ledger_id,
            filter=filter_pb,
            begin=self.conn.codec.encode_begin_offset(begin_offset),
        )

        self._response_stream = response_stream = stub.GetTransactions(request)
        async for response in response_stream:
            for tx in response.transactions:
                for event in tx.events:
                    event_type = event.WhichOneof("event")
                    if event_type == "created":
                        yield await self.conn.codec.decode_created_event(event.created)
                    elif event_type == "archived":
                        yield await self.conn.codec.decode_archived_event(event.archived)
                    else:
                        warnings.warn(f"Unknown Event({event_type}=...)", ProtocolWarning)
                yield Boundary(tx.offset)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains utilities for deterministically outputting contract information.
"""

from typing import TYPE_CHECKING, Optional, TextIO

from .fmt_base import get_formatter
from .model import TableBuilder

if TYPE_CHECKING:
    from ...client import Network

__all__ = ["write_acs"]


def write_acs(
    buf: "TextIO", network: "Network", fmt: "Optional[str]" = None, include_archived: bool = False
) -> None:
    """
    Write all entries tabled by the plugin to the provided buffer.

    :param buf:
        The buffer to write to.
    :param network:
        The network to write data from.
    :param fmt:
        Format of the output; currently either ``"json"`` or ``"pretty"``.
    :param include_archived:
        ``True`` to include archived contracts in the output; otherwise, ``False``.
    """
    formatter = get_formatter(fmt)
    parties = set(network.parties())

    table = TableBuilder()
    for party in parties:
        client = network.aio_party(party)

        if include_archived:
            for cxdata in client.find_historical("*"):
                table.add(party, cxdata.cid, cxdata.cdata, cxdata.effective_at)
        else:
            for cid, cdata in client.find_active("*").items():
                table.add(party, cid, cdata, None)

    for line in formatter.render(network.lookup, parties, table):
        buf.write(line + "\n")



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Contains functions for working with "native" Python types as they correspond to types over the
Ledger API.
"""

from typing import Any, Dict, Mapping, Tuple

__all__ = ["to_record", "to_variant"]

VARIANT_KEYS = frozenset(["tag", "value"])


def to_record(obj: "Any") -> "Mapping[str, Any]":
    """
    "Unflattens" fields of a dict to support nested records.
    """
    from collections.abc import Mapping

    if not isinstance(obj, Mapping):
        raise ValueError("a mapping is required")

    # pull out any specialized dotted-field mappings
    reformatted = dict()  # type: Dict[str, Any]
    for key, value in obj.items():
        k1, d, k2 = key.partition(".")
        if d:
            v = reformatted.get(k1)
            if v is None:
                v = dict()
                reformatted[k1] = v
            v[k2] = value
        else:
            reformatted[key] = value

    return reformatted


def to_variant(obj: "Any") -> "Tuple[str, Any]":
    """
    Return the constructor and value that is represented by the given object.
    """
    from collections.abc import Mapping

    if not isinstance(obj, Mapping):
        raise ValueError(f"cannot coerce {obj!r} to a variant")

    if VARIANT_KEYS == obj.keys():
        return obj["tag"], obj["value"]
    if len(obj) != 1:
        raise ValueError(f"variants must be encoded as single-key dicts (got {obj!r} instead)")
    key = list(obj)[0]
    value = obj.get(key)
    return key, value
